# -*- coding: utf-8 -*-

__license__ = "GNU General Public License (GPL) Version 3"
__author__ = 'Dan McDougall <YouKnowWho@YouKnowWhat.com>'

from plugin_utils import get_translation
from configfile import ConfigFile

# Take care of internationalization
trans = get_translation(__file__)
_ = lambda s: trans.ugettext(s)

class Group(object):
    """Class for referencing and saving group properties (GID and users)"""

    class GroupException(Exception):
        """A custom exception for reporting problems performing operations within Group()"""
        def __init__(self, value):
            self.parameter = value
        def __str__(self):
            return repr(self.parameter)

    def __init__(self, groupname, password, gid, users=None):
        """Initialize and assign group properties."""
        self.groupname = groupname
        self.password = password # Yes, I know this is never used
        self.gid = gid
        if isinstance(users, list):
            self.users = users[:] # Make a copy of it so we can check for changes later
        else: # users is a string (likely a comma-separated list)
            if ',' in users:
                self.users = users.split(',')
            else:
                self.users = users

    def __str__(self):
        self_asList = self.asList()[:] # Make a copy
        if len(self_asList) > 3:
            if isinstance(self_asList[3], list):
                self_asList[3] = ",".join(self_asList[3])
        return ":".join(self_asList)

    def asList(self):
        """Returns the group's parameters as a list (identical to  how it would be represented in self.group_obj.asList)"""
        return [self.groupname, self.password, self.gid, self.users]

    def save(self):
        """Saves/updates the group's properties in /etc/group"""
        groupfile = GroupFile()
        updating = False # This is used when we are making an update to an existing record (as opposed to adding a new one)
        for index, groupobj in enumerate(groupfile.asList):
            if groupobj.groupname == self.groupname:
                groupfile.asList[index] = self
                updating = True
                break
        if updating: # We're done
            groupfile.save()
        else: # We're appending
            groupfile.asList.append(self)
            groupfile.save()

    def delete(self):
        """Permanently delete the group from /etc/group."""
        groupfile = GroupFile()
        for groupobj in groupfile:
            if groupobj.groupname == self.groupname:
                groupfile.remove(groupobj)
                break
        groupfile.save()

    def __iter__(self):
        return self.asList().__iter__()

    def __len__(self):
        return len(self.asList())

    def __contains__(self, val):
        for item in self.asList():
            if item == val:
                return True
        else:
            return False

    def __getitem__(self, val):
        """This lets you access Group objects as a dictionary or a list (just like User() objects)."""
        try:
            return self.__dict__[val]
        except:
            try:
                return self.asList()[val]
            except:
                return None

class GroupFile(ConfigFile):
    """ConfigFile class to read and write to /etc/group"""

    def read_group_file(self):
        """Read /etc/group and returns the data as a list of Group() objects."""
        out_list = []
        for line in open(self.filepath, 'r').readlines():
            fields = line.rstrip('\n').split(':')
            groupobj = Group(fields[0], fields[1], fields[2], fields[3].split(','))
            out_list.append(groupobj)
        return out_list

    def __init__(self):
        """Sets up some basic properties and executes self.read_group_file()."""
        self.filepath = '/etc/group'
        self.asList = self.read_group_file()

    def __str__(self):
        out_string = ""
        for groupobj in self.asList:
            out_string += str(groupobj) + '\n'
        return out_string.rstrip('\n')

    def __repr__(self):
        return str(self.asList)

    def __iter__(self):
        return self.asList.__iter__()

    def __len__(self):
        return len(self.asList)

    def __contains__(self, group):
        """This let's you do things like this:
            groupfile = GroupFile()
            if 'somegroup' in groupfile:
                print "someuser is present in /etc/group"
        ...and things like this:
            somegroup = Group('testgroup', 'x', '1111', ['testuser', 'someuser'])
            if somegroup in groupfile:
                print "someuser is present in /etc/group"

        Note: This obviously is only designed to work with group names
        """
        if isinstance(group, str):
            for groupobj in self.asList:
                if groupobj.groupname == group:
                    return True
        elif isinstance(group, Group):
            for groupobj in self.asList:
                if groupobj.groupname == group.groupname:
                    return True
        else:
            return False

    def __getitem__(self, group):
        """This lets you access GroupFile as a dictionary (of users).

        Example:
            groupfile = GroupFile()
            print groupfile['testgroup']
            testgroup:x:1111:testuser"""
        if isinstance(group, str):
            for groupobj in self.asList:
                if groupobj.groupname == group:
                    return groupobj
        elif isinstance(group, Group):
            for groupobj in self.asList:
                if groupobj.groupname == group.groupname:
                    return groupobj

    def groups(self):
        """Returns a list of groups in /etc/group"""
        out_list = []
        for line in self.asList:
            out_list.append(line[0])
        return out_list

    def gids(self):
        """Returns a list of gids in /etc/group"""
        out_list = []
        for line in self.asList:
            out_list.append(line[2])
        return out_list

    def group(self, group_name):
        """If the group exists, returns a Group object generated from said group's properties in /etc/group."""
        for group in self.asList:
            if group.groupname == group_name:
                groupobj = Group(group[0], group[1], group[2], group[3])
                return groupobj

    def save(self):
        """Overwrites /etc/group with the data from self.asList()."""
        group_ondisk = GroupFile() # Create a copy of ourselves from the on-disk files
        if str(self) == str(group_ondisk): # Check for changes from /etc/group
            return _("No changes to save in /etc/group")
        else:
            # Convert each Group object's user list into a comma-separated string (for save_delimited)
            for group in self.asList:
                if isinstance(group.users, list): # This leaves already-comma-separated strings alone
                    group.users = ",".join(group.users)
            self.save_delimited(self.asList, '/etc/group', ':')
        self.asList = self.read_group_file()

    def remove(self, groupobj):
        """Meant to emulate the remove() built-in (for lists)."""
        self.asList.remove(groupobj)

def list_etc_group():
    """Returns /etc/group as a list (skipping the password placeholder field)"""
    group = open('/etc/group').readlines()
    group_list = []
    for line in group:
        line = line.rstrip('\n')
        line_list = line.split(':')
        line_list.pop(1) # Drop the password placeholder field
        line_list[1] = int(line_list[1])
        group_list.append(line_list)
    return group_list