# -*- coding: utf-8 -*-
from commands import getoutput

__license__ = "GNU General Public License (GPL) Version 3"
__author__ = 'Dan McDougall <YouKnowWho@YouKnowWhat.com>'
__version__ = '1.0'
__version_info__ = (1, 0)

__doc__ == """
iproute2.py: A module that can be used to get the status of all interfaces
on the current host using the 'ip' command (iproute2).
"""

def keyword_value(keyword, text):
    """
    Scans 'text' and returns the word immediately following 'keyword'.
    Example:
        >>>text = "1: lo    inet 127.0.0.1/8 brd 127.255.255.255 scope host lo"
        >>>keyword_value('scope', text)
        'host'
    """
    split_text = text.split()
    for index, word in enumerate(split_text):
        if word == keyword:
            break # We just need the index
    try:
        return split_text[index + 1]
    except:
        return 'N/A'

class Inet(object):
    """
    A class to store IPv4 addresses and related info (i.e. broadcast, scope)
    """
    def __init__(self, iface, inet, broadcast, scope):
        self.iface = iface
        self.inet = inet
        self.broadcast = broadcast
        self.scope = scope

    def __repr__(self):
        out_dict = {
            'inet': self.inet,
            'broadcast': self.broadcast,
            'scope': self.scope
        }
        return str(out_dict)

    def __getitem__(self, val):
        return self.__dict__[val]

class Inet6(object):
    """
    A class to store IPv6 addresses and related info (i.e. broadcast, scope)
    """
    def __init__(self, iface, inet6, scope, valid_lft, preferred_lft):
        self.iface = iface
        self.inet6 = inet6
        self.scope = scope
        self.valid_lft = valid_lft
        self.preferred_lft = preferred_lft

    def __repr__(self):
        out_dict = {
            'inet6': self.inet6,
            'scope': self.scope,
            'valid_lft': self.valid_lft,
            'preferred_lft': self.preferred_lft
        }
        return str(out_dict)

    def __getitem__(self, val):
        return self.__dict__[val]

    def __contains__(self, val):
        for item in self.__dict__:
            if item == val:
                return True
        else:
            return False

class IP(object):
    """
    A class to represent the output of the 'ip' command.
    """
    def __init__(self):
        self.interfaces = {}
        self.parse_ip_command()


    def parse_ip_command(self):
        """
        Uses the 'ip' command to assemble the self.interfaces dict.

        self.interfaces is laid out like so:
            {
                'lo': {
                    'num': 1,
                    'mtu': 1500,
                    'qdisc': 'noqueue',
                    'qlen': None,
                    'state': 'UNKNOWN',
                    'link': 'loopback',
                    'mac': '00:00:00:00:00:00',
                    'inets': ['127.0.0.1/8'],
                    'inet6s': ['::1/128'],
                    'flags': 'valid_lft forever preferred_lft forever'
                }
            }

        """
        ifnames_list = []
        output = getoutput('ip -o addr')
        # Note: I'm sure there's a regex that could split 'output' and simplify this quite a bit
        #       ...anyone care to give it a go?
        split_output = output.split('\n')
        for line in split_output:
            # First collect all the interface names
            ifname = line.split()[1].rstrip(':')
            if ifname not in ifnames_list:
                ifnames_list.append(ifname)
        # Now assemble self.interfaces
        map(self.ip_addr_info, ifnames_list)

    def ip_addr_info(self, ifname):
        """
        Parses 'ip addr show dev <iface>' and updates self.interfaces with the result
        """
        link = None
        qdisc = None # Queuing Discipline
        qlen = None # Queue Length
        mac = None
        inets = []
        inet6s = []
        flags = []
        broadcast = None
        valid_lft = None
        preferred_lft = None

# TODO:  Finish changing flags to qflags and write the flags bit (i.e. MULTICAST,UP,NO-CARRIER)
        # Grab the easy ones
        output = getoutput('ip addr show dev %s' % ifname)
        num = output[0]
        mtu = keyword_value('mtu', output)
        qdisc = keyword_value('qdisc', output)
        state = keyword_value('state', output)

        # Now for the harder ones...
        # Assemble the flags...
        temp = [a for a in output.split('\n') if 'mtu ' in a]
        if temp:
            for line in temp:
                flags = line.split()[2].lstrip('<').rstrip('>').split(',')

        # Assemble IPv4 objects...
        temp = [a for a in output.split('\n') if 'inet ' in a]
        if temp:
            for inet_line in temp:
                inet = keyword_value('inet', inet_line)
                if 'brd' in inet_line:
                    broadcast = keyword_value('brd', inet_line)
                scope = keyword_value('scope', temp[0])
                inets.append(Inet(ifname, inet, broadcast, scope))

        # Assemble IPv6 objects...
        temp = [a for a in output.split('\n') if 'inet6' in a]
        if temp:
            for index, inet_line in enumerate(output.split('\n')):
                if inet_line.lstrip().startswith('inet6'):
                    # Join the next line to this one since they belong together
                    # ...like peanut butter and chocolate...  Mmmm
                    inet6_joined = " ".join([
                        output.split('\n')[index].strip(),
                        output.split('\n')[index + 1].strip()
                    ])
                    # So inet6_joined should look something like this now:
        # 'inet6 fe80::250:56ff:fec0:8/64 scope link valid_lft forever preferred_lft forever'
                    inet = keyword_value('inet6', inet6_joined)
                    scope = keyword_value('scope', inet6_joined)
                    valid_lft = keyword_value('valid_lft', inet6_joined)
                    preferred_lft = keyword_value('preferred_lft', inet6_joined)
                    inet6s.append(Inet6(ifname, inet, scope, valid_lft, preferred_lft))

        # Grab the queue length (if set)
        try:
            qlen = keyword_value('qlen', output)
        except IndexError:
            pass # No qlen set

        # Grab the link type
        temp = [a for a in output.split('\n') if 'link/' in a]
        if temp:
            link = temp[0].split()[0].split('/')[1]
            try:
                mac = temp[0].split()[1]
            except IndexError:
                pass # Not everything has a MAC

        # Now put it all together
        self.interfaces.update({
            ifname: {
                'num': num,
                'inets': inets,
                'inet6s': inet6s,
                'mtu': mtu,
                'qdisc': qdisc,
                'state': state,
                'qlen': qlen,
                'mac': mac,
                'link': link,
                'flags': flags,
            }
        })