# -*- coding: utf-8 -*-

__license__ = "GNU General Public License (GPL) Version 3"
__author__ = 'Dan McDougall <YouKnowWho@YouKnowWhat.com>'

import os, re, gettext
from utils import format_columns
from configfile import ConfigFile
from odict import odict

__doc__ = "Coova Chilli configuration file library"

locale_dir = os.path.join(os.getcwd(), 'i18n')
def get_translation():
    try:
        trans = gettext.translation('PyCI', locale_dir)
    except:
        trans = gettext.NullTranslations()
    return trans
trans = get_translation()
_ = lambda s: trans.ugettext(s).encode('utf-8')

class CoovaChilliConfig(ConfigFile):
    """
    Class to store and save Coova Chilli configuration files.
    """
    settings_list = [
        "HS_ADMPWD",
        "HS_ADMUSR",
        "HS_ANYIP",
        "HS_DEFBANDWIDTHMAXDOWN",
        "HS_DEFBANDWIDTHMAXUP",
        "HS_DEFIDLETIMEOUT",
        "HS_DEFSESSIONTIMEOUT",
        "HS_DNS1",
        "HS_DNS2",
        "HS_DNS_DOMAIN",
        "HS_DNSPARANOIA",
        "HS_DYNIP",
        "HS_DYNIP_MASK",
        "HS_LANIF",
        "HS_LOC_AC",
        "HS_LOC_CC",
        "HS_LOC_ISOCC",
        "HS_LOC_NAME",
        "HS_LOC_NETWORK",
        "HS_MACALLOW",
        "HS_MACAUTH",
        "HS_MACAUTHDENY",
        "HS_MACAUTHMODE",
        "HS_MODE",
        "HS_NASID",
        "HS_NASIP",
        "HS_NASMAC",
        "HS_NETMASK",
        "HS_NETWORK",
        "HS_OPENIDAUTH",
        "HS_POSTAUTH_PROXY",
        "HS_POSTAUTH_PROXYPORT",
        "HS_PROVIDER",
        "HS_PROVIDER_LINK",
        "HS_RADACCT",
        "HS_RADAUTH",
        "HS_RADCONF",
        "HS_RADCONF_AUTHPORT",
        "HS_RADCONF_PWD",
        "HS_RADCONF_SECRET",
        "HS_RADCONF_SERVER",
        "HS_RADCONF_URL",
        "HS_RADCONF_USER",
        "HS_RADIUS",
        "HS_RADIUS2",
        "HS_RADSECRET",
        "HS_SSID",
        "HS_STATIP",
        "HS_STATIP_MASK",
        "HS_TYPE",
        "HS_UAMALLOW",
        "HS_UAMDOMAINS",
        "HS_UAMFORMAT",
        "HS_UAMHOMEPAGE",
        "HS_UAMLISTEN",
        "HS_UAMPORT",
        "HS_UAMSECRET",
        "HS_UAMSERVER",
        "HS_UAMSERVICE",
        "HS_USELOCALUSERS",
        "HS_USE_MAP",
        "HS_WANIF",
        "HS_WPAGUESTS",
        "HS_WWWBIN",
        "HS_WWWDIR"
    ]

# TODO:  Finish filling these out
    settings_help = {
        'HS_SSID': "The SSID to send to the captive portal.",
        'HS_NETWORK': "HotSpot network (must include HS_UAMLISTEN)",
        'HS_NETMASK': "HotSpot network netmask",
        'HS_UAMLISTEN': "HotSpot IP Address (on subscriber network)",
        'HS_UAMSERVER': "The server to be used in combination with HS_UAMFORMAT to create the final chilli 'uamserver' url configuration.",
        'HS_UAMFORMAT': "Use HS_UAMFORMAT to define the actual captive portal url.",
        'HS_UAMSERVICE': 'This option will be configured to be the WISPr LoginURL as well as provide "uamService" to the ChilliController.',
        'HS_UAMSECRET': "The password the UAM server uses in conjunction with the chilli daemon to prevent enumeration attacks.",
        'HS_RADIUS': "Primary radius server to use for authentication, accounting, and authorization (AAA).",
        'HS_RADIUS2': "Secondary (fallback) radius server to use for authentication, accounting, and authorization (AAA).",
        'HS_RADSECRET': "Shared secret to use with the radius server(s)",
        'HS_UAMALLOW': "Hosts and networks (e.g. '10.1.0.0/24') that users can access without requiring authentication.  With this setting you can create a 'walled garden'.",
        'HS_UAMDOMAINS': "Wildcard domain names that users are allowed to access without requiring authentication (e.g. '.paypal.com').",
        'HS_ADMUSR': "Admin username",
        'HS_ADMPWD': "Admin password"
    }

# These are to make things easier for the user
    settings_categories = odict([
        ('local_network_settings', [
            'HS_LANIF',
            'HS_NETWORK',
            'HS_NETMASK',
            'HS_UAMLISTEN',
            'HS_UAMPORT',
            'HS_DYNIP',
            'HS_DYNIP_MASK',
            'HS_STATIP',
            'HS_STATIP_MASK',
            'HS_DNS_DOMAIN',
            'HS_DNS1',
            'HS_DNS2',
            'HS_DNSPARANOIA',
        ]),
        ('hotspot_settings', [
            'HS_NASID',
            'HS_UAMSECRET',
            'HS_RADIUS',
            'HS_RADIUS2',
            'HS_RADSECRET',
            'HS_UAMALLOW',
            'HS_UAMDOMAINS',
            'HS_SSID',
            'HS_NASMAC',
            'HS_NASIP',
            'HS_UAMSERVER',
            'HS_UAMFORMAT',
            'HS_UAMHOMEPAGE',
            'HS_UAMSERVICE',
            'HS_WWWDIR',
            'HS_WWWBIN',
            'HS_PROVIDER',
            'HS_PROVIDER_LINK',
            'HS_DEFSESSIONTIMEOUT',
            'HS_DEFIDLETIMEOUT',
            'HS_ANYIP',
            'HS_MACAUTH',
            'HS_MACAUTHDENY',
            'HS_MACAUTHMODE',
            'HS_MACALLOW',
            'HS_USELOCALUSERS',
            'HS_OPENIDAUTH',
            'HS_WPAGUESTS',
            'HS_OPENIDAUTH',
            'HS_USE_MAP',
            'HS_POSTAUTH_PROXY',
            'HS_POSTAUTH_PROXYPORT'
        ]),
        ('wispr_settings', [
            'HS_LOC_NAME',
            'HS_LOC_NETWORK',
            'HS_LOC_AC',
            'HS_LOC_CC',
            'HS_LOC_ISOCC',
            'HS_DEFBANDWIDTHMAXDOWN',
            'HS_DEFBANDWIDTHMAXUP',
        ]),
        ('centralized_config_settings', [
            'HS_RADCONF',
            'HS_RADCONF_URL',
            'HS_RADCONF=on',
            'HS_RADCONF_SERVER',
            'HS_RADCONF_SECRET',
            'HS_RADCONF_AUTHPORT',
            'HS_RADCONF_USER',
            'HS_RADCONF_PWD',
        ]),
    ])

    def __init__(self, filepath="/etc/chilli/config"):
        """
        Read a Coova Chilli config file into a dictionary that can be
        saved via the standard ConfigFile methods.
        """
        self.filepath = filepath
        if self.filepath:
            self.settings = self.read_config_file()

    def read_config_file(self, filepath=None):
        """Reads filepath (default: self.filepath) and returns a dictionary
        of found configuration parameters."""
        settings_dict = odict()
        single_quoted_regex = re.compile('\'(.*?)\'', re.IGNORECASE) # This matches things 'in quotes'
        double_quoted_regex = re.compile('"(.*?)"', re.IGNORECASE) # This matches things "in quotes"
        if not filepath:
            filepath = self.filepath
        if os.path.exists(filepath):
            for line in open(filepath, 'r'):
                line = line.strip()
                if line.startswith('#'): # Comment
                    pass # ignore
                elif '=' in line: # There's a setting here
                    if '"' in line: # The option is set to something in double quotes
                        line = double_quoted_regex.split(line)
                        option = line[0].rstrip('=')
                        value = line[1]
                    elif "'" in line: # The option is set to something in double quotes
                        line = single_quoted_regex.split(line)
                        option = line[0].rstrip('=')
                        value = line[1]
                    else: # Just 'OPTION=VALUE'
                        line = line.split()[0] # This gets rid of any comments or whatnot after the option
                        line = line.split('=')
                        option = line[0]
                        value = line[1]
                    settings_dict.update({option: value})
        return settings_dict

    def empty_config(self):
        """Returns an empty odict object in the event we're working with an empty or non-existant file."""
        empty_conf_object = odict()
        return empty_conf_object

    def save(self):
        """
        Saves the config file (does not preserve comments)
        """
        # TODO: Write some validation
        # TODO:  Make this preserve comments (idea: go through the file line-by-line and replace each option as necessary)
        #self.validate()
        self.make_backup(self.filepath)
        f = open(self.filepath, 'w')
        f.write(str(self))
        f.close()

    def save_raw(self, text):
        """Saves the raw (plaintext) version of the config file using 'text'.
        This overrides the default save_raw function in since we don't need to run self.made_changes()."""
        self.make_backup(self.filepath)
        f = open(self.filepath, 'w')
        f.write(text)
        f.close()

    def __str__(self):
        """Returns a string representing how the config file will look on the filesystem."""
        out_string = ""
        for option, value in self.settings.items():
            # Note: this automatically puts everything in double-quotes
            out_string += '%s="%s"\n' % (option, value)
        return out_string

def coova_chilli_status():
    """Returns a list of lists representing the output of the 'chilli_query list' command."""
    from commands import getoutput
    command = "/usr/sbin/chilli_query list"
    output = getoutput(command)
    out_list = []
    for line in output.splitlines():
        out_list.append(line.split()) # Nice, space-separated columns =)
    return out_list

def authorize_user(sessionid, **kwargs):
    """Authorizes the given user (sessionid) with the provided keyword arguments
    using the chilli_query command (e.g. 'chilli_query authorize sessionid <sessionid> ...').
    kwargs is expected to be something like this:
        {
            'username': 'jdoe',
            'maxbwup': '12',
            'maxinputoctects': '12',
            'maxoutputoctects': '12',
            'idletimeout': '12',
            'maxoctects': '12',
            'sessiontimeout': '12',
            'maxbwdown': '12'
        }
    """
    from commands import getstatusoutput
    command = "/usr/sbin/chilli_query authorize sessionid %s" % sessionid
    for key, value in kwargs.items():
        if value:
            command += " %s %s" % (key, value)
    #print "command: %s" % command
    retcode, result = getstatusoutput(command)
    # chilli_query authorize doesn't give any output (even if the sessionid doesn't exist)
    # ...and even if it doesn't exist it still returns 0
    return retcode

def logout_user(mac_address):
    """Executes 'chilli_query logout <mac address>'"""
    from commands import getstatusoutput
    command = "/usr/sbin/chilli_query logout %s" % mac_address
    #print "command: %s" % command
    retcode, result = getstatusoutput(command)
    return retcode