# -*- coding: utf-8 -*-
import sqlite3
from utils import get_groups

__license__ = "GNU General Public License (GPL) Version 3"
__author__ = 'Dan McDougall <YouKnowWho@YouKnowWhat.com>'

def connect_db(db):
    """Connect to the sqlite3 database path, 'db' and return it as an sqlite3 object"""
    return sqlite3.connect(db,  isolation_level=None)

def disconnect_db(db):
    """close() the given sqlite3 object"""
    db.close()

def delete_user_permission_by_id(key):
    """Deletes the row where id='key' in 'db'"""
    pyci_db = connect_db('pyci.db')
    print "key: %s" % key
    dbcursor = pyci_db.cursor()
    dbcursor.execute("DELETE FROM user_permissions WHERE id=?", (key,))
    return True

def delete_group_permission_by_id(key):
    """Deletes the row where id='key' in 'db'"""
    pyci_db = connect_db('pyci.db')
    dbcursor = pyci_db.cursor()
    dbcursor.execute("DELETE FROM group_permissions WHERE id=(?)", (key,))
    return True

def change_user_permission(user, plugin, permission, id=None):
    """Modifies a user's permissions to a given plugin. 'permission' should either be 'r' or 'rw'."""
    pyci_db = connect_db('pyci.db')
    dbcursor = pyci_db.cursor()
    if permission == None:
        # Remove the existing permission
        dbcursor.execute("DELETE FROM user_permissions WHERE user=(?) AND plugin=(?)", (user, plugin))
        disconnect_db(pyci_db)
    elif id:
        # Update the existing permission
        dbcursor.execute("UPDATE user_permissions SET user=(?), plugin=(?), permission=(?) WHERE id=(?)", (user, plugin, permission, id))
        disconnect_db(pyci_db)
    else:
        # Create a new permission
        dbcursor.execute("INSERT INTO user_permissions VALUES (null, ?, ?, ?)", (user, plugin, permission))
        disconnect_db(pyci_db)
    return True

def change_group_permission(group, plugin, permission, id=None):
    """Adds or modifies a group's permissions to a given plugin. permission should either be 'r' or 'rw'."""
    pyci_db = connect_db('pyci.db')
    dbcursor = pyci_db.cursor()
    if permission == None:
        # Remove the existing permission
        dbcursor.execute("DELETE FROM group_permissions WHERE _group=(?) AND plugin=(?)", (group, plugin))
        disconnect_db(pyci_db)
    elif id:
        # Update the existing permission
        dbcursor.execute("UPDATE group_permissions SET _group=(?), plugin=(?), permission=(?) WHERE id=(?)", (group, plugin, permission, id))
        disconnect_db(pyci_db)
    else:
        # Create a new permission
        dbcursor.execute("INSERT INTO group_permissions VALUES (null, ?, ?, ?)", (group, plugin, permission))
        disconnect_db(pyci_db)
    return True

def check_user_permission(user=None, plugin=None, orderby=None, sortorder=None):
    """Returns a user's permissions for a given plugin ('r' or 'rw').
    If no user or plugin are supplied, return a list of all user permissions.
    If only a user is supplied, return a list of all their permissions.
    If only a plugin is supplied, return a list of all related user permissions."""
    pyci_db = connect_db('pyci.db')
    dbcursor = pyci_db.cursor()
    if not sortorder:
        sortorder = "DESC"
    if not orderby:
        ordered_select = "SELECT id, user, plugin, permission FROM user_permissions ORDER BY id"
    else: # Note: I had to use %s below because "ORDER BY ?" doesn't work for some reasons
        ordered_select = "SELECT id, user, plugin, permission FROM user_permissions ORDER BY %s %s" % (orderby.lower(), sortorder)
    if not user and not plugin:
        permission_list = dbcursor.execute(ordered_select).fetchall()
        disconnect_db(pyci_db)
        return permission_list
    if user and not plugin:
        permission_list = []
        for row in dbcursor.execute("SELECT id, user, plugin, permission FROM user_permissions WHERE user=(?)", (user,)):
            permission_list.append(row)
        disconnect_db(pyci_db)
        return permission_list
    if not user and plugin:
        permission_list = []
        for row in dbcursor.execute("SELECT id, user, plugin, permission FROM user_permissions WHERE plugin=(?)", (plugin,)):
            permission_list.append(row)
        disconnect_db(pyci_db)
        return permission_list
    dbcursor.execute("SELECT id, user, plugin, permission FROM user_permissions WHERE (user=(?) AND plugin=(?)) OR (user=(?) AND plugin='all')", (user, plugin, user))
    raw_row = dbcursor.fetchone()
    if raw_row:
        # raw_row should be a tuple that looks like this:
        # (1, u'someuser', u'status', u'r')
        disconnect_db(pyci_db)
        return raw_row[3] # Return the actual permission ('r' or 'rw')
    else:
        # The user doesn't have permission so now we check if they are in a group that has permission
        permission_list = []
        for group in get_groups(user):
            dbcursor.execute("SELECT id, _group, plugin, permission FROM group_permissions WHERE (_group=(?) AND plugin=(?)) OR (_group=(?) AND plugin='all')", (group, plugin, group))
            raw_row = dbcursor.fetchone()
            if raw_row:
                permission_list.append(raw_row[2])
        disconnect_db(pyci_db)
        if 'rw' in permission_list:
            return 'rw'
        elif 'r' in permission_list:
            return 'r'
        else:
            return None

def check_group_permission(group=None, plugin=None, orderby=None, sortorder=None):
    """Returns a group's permissions for a given plugin ('r' or 'rw').
    If no group or plugin are supplied, return a list of all group permissions.
    If only a group is supplied, return a list of all its permissions.
    If only a plugin is supplied, return a list of all related group permissions."""
    pyci_db = connect_db('pyci.db')
    dbcursor = pyci_db.cursor()
    if not sortorder:
        sortorder = "DESC"
    if not orderby:
        ordered_select = "SELECT id, _group, plugin, permission FROM group_permissions ORDER BY id"
    else: # Note: I had to use %s below because "ORDER BY ?" doesn't work for some reasons
        ordered_select = "SELECT id, _group, plugin, permission FROM group_permissions ORDER BY %s %s" % (orderby.lower(), sortorder)
    if not group and not plugin:
        permission_list = dbcursor.execute(ordered_select).fetchall()
        disconnect_db(pyci_db)
        return permission_list
    if group and not plugin:
        permission_list = []
        for row in dbcursor.execute("SELECT id, _group, plugin, permission FROM group_permissions WHERE _group=(?)", (group,)):
            permission_list.append(row)
        return permission_list
    if not group and plugin:
        permission_list = []
        for row in dbcursor.execute("SELECT id, _group, plugin, permission FROM group_permissions WHERE plugin=(?)", (plugin,)):
            permission_list.append(row)
        return permission_list
    dbcursor.execute("SELECT id, _group, plugin, permission FROM group_permissions WHERE (_group=(?) AND plugin=(?)) OR (_group=(?) AND plugin='all')", (group, plugin, group))
    raw_row = dbcursor.fetchone()
    if raw_row:
        disconnect_db(pyci_db)
        return raw_row[3] # Return the actual permission ('r' or 'rw')
    else:
        return None

def cleanup_permissions(user):
    """If a user has been granted 'all' permissions they only need one entry in the permissions table.
    This function gets rid of all the extraneous entries and just leaves one 'all' entry"""
    # TODO:  Write this
    pass