# -*- coding: utf-8 -*-

__license__ = "GNU General Public License (GPL) Version 3"
__author__ = 'Dan McDougall <YouKnowWho@YouKnowWhat.com>'

# Form based authentication for CherryPy. Requires the
# Session tool to be loaded.

# Most of this code taken from http://tools.cherrypy.org/wiki/AuthenticationAndAccessRestrictions
# Thank you Arnar Birgisson and "Ivan"!

import re, os
from commands import getstatusoutput # Note:  In Python 3 getstatusoutput has been moved to the subprocess module
import cherrypy
from cherrypy.lib import httpauth
import urllib
from pyci_db import check_user_permission, check_group_permission, connect_db, disconnect_db
from utils import lookup_template

# Internationalization
import gettext
locale_dir = os.path.join(os.getcwd(), 'i18n')
def get_translation():
    try:
        trans = gettext.translation(domain='PyCI', localedir=locale_dir)
        #trans.install(unicode=True)
    except: # Use a null translation (i.e. English) if no language is found matching the current LANG environment variable
        trans = gettext.NullTranslations()
        #trans.install(unicode=True)
    return trans
trans = get_translation()
_ = lambda s: trans.ugettext(s).encode('utf-8') # The encode() fixes UnicodeDecodeError exceptions on the console

SESSION_KEY = '_cp_username'

def get_group_memeberships(user):
    """Returns a list of groups the user belongs to (including their primary group)"""
    # "id -G -n" returns all groups a user belongs to including their primary group (despite what the help says)
    group_memberships = getstatusoutput("id -G -n %s" % user)[1]
    return group_memberships.split()

def check_credentials(user, password):
    """Checks a given user and password against /etc/shadow (or /etc/passwd if /etc/shadow doesn't exist).
    Returns None on success and an error message on failure."""
    from crypt import crypt
    shadow_hash = ''
    salt_regex = re.compile(r'\$.*\$.*\$')
    if os.path.exists('/etc/shadow'):
        password_file = '/etc/shadow'
    else:
        password_file = '/etc/passwd'
    shadow = open(password_file, 'r').readlines()
    for line in shadow:
        cols = line.split(':')
        if cols[0] == user:
            shadow_hash = cols[1]
    if salt_regex.match(shadow_hash):
        salt = salt_regex.match(shadow_hash).group()
        hashed_pass = crypt(password, salt)
        if hashed_pass == shadow_hash:
            return None
    # Now check the case where OpenWRT was just started up for the first time and there's no root password set yet...
    # If the root password hasn't been set the telnet daemon will be running with /bin/login.sh (i.e. no-password login)
    retcode = getstatusoutput('ps aux | grep -v grep | grep "telnetd -l /bin/login.sh"')[0]
    re.purge()
    if retcode == 0: # A return code of 0 means grep actually grepped something (as opposed to no output)
        return None # Let them through
    else:
        return _("Invalid user and/or password")

def check_group_membership(username, group):
    """Checks whether or not user is a member of 'group' (using /etc/passwd and /etc/group)
    Returns True if the user is in the group.  False if not."""
    # First check the group file
    group_regex = re.compile(r'^%s:.*' % group)
    groupfile = open('/etc/group', 'r')
    for line in groupfile:
        if group_regex.match(line):
            fields = line.split(':')
            gid = fields[2]
            group_users = fields[3].strip().split(',')
    if username in group_users: # username is in the required group
        return True
    # Now find the user's primary group GID in /etc/passwd so we can check that
    passwd = open('/etc/passwd', 'r').readlines()
    for line in passwd:
        line = line.split(':')
        if line[0] == username:
            primary_group_gid = line[3]
            break # No need to continue in this loop
    passwd.close() # Always remember to put your toys away
    groupfile.seek(0) # Advance to GO (Collect $200)
    for line in groupfile:
        fields = line.split(':')
        group_name = fields[0]
        group_gid = fields[2]
        if group_gid == primary_group_gid: # Just the group we've been looking for
            if group_name == group: # The user *is* in 'group' after all
                return True
    return False

def check_auth(*args, **kwargs):
    """A tool that looks in config for 'auth.require'. If found and it
    is not None, a login is required and the entry is evaluated as a list of
    conditions that the user must fulfill"""
    conditions = cherrypy.request.config.get('auth.require', None)
    # format GET params
    get_parmas = urllib.quote(cherrypy.request.request_line.split()[1])
    if conditions is not None:
        user = cherrypy.session.get(SESSION_KEY)
        if user:
            cherrypy.request.login = user
            for condition in conditions:
                # A condition is just a callable that returns true orfalse
                if not condition():
                    raise cherrypy.HTTPError(401, "You are not authorized to access that resource")
                    # Send old page as from_page parameter
                    #raise cherrypy.HTTPRedirect("/auth/login?from_page=%s" % get_parmas)
        else:
            # Send old page as from_page parameter
            raise cherrypy.HTTPRedirect("/auth/login?from_page=%s" %get_parmas)

cherrypy.tools.auth = cherrypy.Tool('before_handler', check_auth)

def require(*conditions):
    """A decorator that appends conditions to the auth.require config variable."""
    def decorate(f):
        if not hasattr(f, '_cp_config'):
            f._cp_config = dict()
        if 'auth.require' not in f._cp_config:
            f._cp_config['auth.require'] = []
        f._cp_config['auth.require'].extend(conditions)
        return f
    return decorate


# Conditions are callables that return True
# if the user fulfills the conditions they define, False otherwise
#
# They can access the current user as cherrypy.request.login
#
# Define those at will however suits the application.

def member_of(group):
    def check():
        return check_group_membership(cherrypy.request.login, group)
    return check

def name_is(reqd_user):
    return lambda: reqd_user == cherrypy.request.login

def any_of(*conditions):
    """Returns True if any of the conditions match"""
    def check():
        for c in conditions:
            if c():
                return True
        return False
    return check

def check_permission(user, permission, plugin):
    """Returns true if the given user has the given permission to the given plugin"""
    actual_permission = check_user_permission(user, plugin)
    if actual_permission:
        if permission in actual_permission:
            return True
    else:
        # Now check if the user belongs to any groups that have permission
        groups = get_group_memeberships(user)
        for group in groups: # walk the list
            actual_permission = check_group_permission(group, plugin)
            if actual_permission:
                if permission in actual_permission:
                    return True
        return False

def has_read(plugin):
    """Decorator that returns True if the user has been granted read permissions to the given plugin"""
    def check():
        return check_permission(cherrypy.request.login, 'r', plugin)
    return check

def has_write(plugin):
    """Decorator that returns True if the user has been granted write permissions to the given plugin"""
    def check():
        return check_permission(cherrypy.request.login, 'rw', plugin)
    return check

# By default all conditions are required, but this might still be
# needed if you want to use it inside of an any_of(...) condition
def all_of(*conditions):
    """Returns True if all of the conditions match"""
    def check():
        for c in conditions:
            if not c():
                return False
        return True
    return check

# Controller to provide login and logout actions
class AuthController(object):

    @cherrypy.expose
    def on_login(self, user):
        """Called on successful login"""
        pass

    @cherrypy.expose
    def on_logout(self, user):
        """Called on logout"""
        pass

    @cherrypy.expose
    def get_loginform(self, user, msg=None, from_page="/"):
        login_template = lookup_template("html", "login_form.mako")
        trans = get_translation()
        _ = lambda s: trans.ugettext(s)
        try:
            language = os.environ['LANG']
        except KeyError: # LANG isn't set
            language = u'en_US' # Fall back/default to US English
        return login_template.render_unicode(language=language, gettext_domain='PyCI', _=_, message=msg, user=user, from_page=from_page)

# TODO:  Implement the "page" option where you can get login() to return the whole page you want (for things like RSS or JSON)
# TODO:  Well, it SHOULD work but it doesn't.  Fix it.
    @cherrypy.expose
    def login(self, user=None, password=None, from_page="/", page=None):
        if user is None or password is None:
            return self.get_loginform("", from_page=from_page)

        error_msg = check_credentials(user, password)
        print error_msg
        if error_msg:
            return self.get_loginform(user, error_msg, from_page)
        else: # Successul login
            cherrypy.session[SESSION_KEY] = cherrypy.request.login = user
            self.on_login(user)
            auth_msg = _("%s authenticated successfully") % user
            cherrypy.log(auth_msg)
            if not page:
                raise cherrypy.HTTPRedirect(from_page or "/")
            else:
                return cherrypy.InternalRedirect(page)

    @cherrypy.expose
    def logout(self, from_page="/"):
        sess = cherrypy.session
        user = sess.get(SESSION_KEY, None)
        sess[SESSION_KEY] = None
        if user:
            cherrypy.request.login = None
            self.on_logout(user)
        raise cherrypy.HTTPRedirect(from_page or "/")