# -*- coding: utf-8 -*-
#
# Description: Example plugin

"""
Example Plugin
==============

This plugin is meant to demonstrate how to write new plugins for PyCI.
It has the following methods:
 - **HelloWorld:** Renders, 'Hello, world!' using a basic Mako template that inherits from the theme's base.mako file.
 - **jqGridExample:** Demonstrates how to use PyCI's jqGrid helper functions.
 - **NotificationExample:** Demonstrates how to use PyCI's user notification functions (pop-up and validation messages).

It is highly recommended that you examine the source code of this plugin and its templates in order to better understand
how PyCI handles plugins.
"""

import cherrypy # Useful for variuos things (more below)
from plugin_utils import Plugin, get_translation # These are essential for all plugins
from auth import require, has_read, has_write # These are necessary for permissions
# plugin_template is (likely) essential for your plugin.  The rest are just useful jQuery/jqGrid abstractions:
from utils import plugin_template, jqgrid_select_options, jqgrid_json, jquery_ui_error_msg, jquery_ui_highlight_msg
from odict import odict # Necessary if you want your plugin menu to appear in a particular order

# Internationalization - This sets up the _() function that you can use with gettext translations.
#   The PyCI build.sh script will look for all _('string here') texts and will automatically generate
#   gettext scaffolds for translators to use.  Just make sure to make a directory named, 'i18n' in your
#   plugin's directory.
trans = get_translation(__file__)
_ = lambda s: trans.ugettext(s)

# This function is for the jqGrid example
def list_etc_hosts():
    """
    Returns /etc/hosts as a list of lists.
    Suitable for things like jqgrid and spreadsheet output
    """
    hostname_list = []
    # Assemble a list of lists of ips and hostnames...
    etc_hosts = open('/etc/hosts')
    for line in etc_hosts:
        if line:
            if line.startswith('#'): # This is a comment
                pass # Skip it
            elif len(line.split()) > 0:
                line = line.split()
                ip = line[0]
                hostnames = line[1:]
                hostname_list.append([ip, " ".join(hostnames)])
    return hostname_list

# All plugins are derived from the Plugin class which is located inside plugin_utils.py
# ...plugins are automatically discovered by PyCI by examining Plugin.__subclasses__
class Example(Plugin):
    """An example plugin."""
    # This plugin is to be displayed in the menu:
    display = True
    # This determines the plugin's order in the menu:
    sort_priority = 1 # 1 is the default priority
    # This parameter represents your plugin's submenu
    navigable_functions = odict([
        # The format is as follows:
        # (_('Title that will end up in your plugin's submenu'), 'ActualMethodName'),
        (_('Hello, World!'), 'HelloWorld'),
        (_('jqGrid Example'), 'jqGridExample'),
        (_('Notification Example'), 'NotificationExample'),
        ])

    # The @require decorator handles the permissions for this method.
    # By default PyCI's permissions interface enumerates plugins and allows
    # users to assign read or read/write permissions to users or groups
    # based on the plugin's name.  In this example, we're requiring that the
    # user have read access to the Example plugin in order to execute (view)
    # this method:
    @require(has_read('Example'))
    def HelloWorld(self):
        """
        Render a basic page using a Mako template that says 'Hello, world!'::

            @require(has_read('Example'))
            def HelloWorld(self):
                page_template = plugin_template(self, "templates/helloworld.mako")
                help=_("<p>This is the example plugin's 'Hello, world!' page.</p>")
                return self.standard_render(
                    template=page_template,
                    help=help,
                    long_name=_('Example Plugin: Hello, world!'),
                    description=_("An example plugin for developers (Hello, world!)."),
                )

        A couple things are going on in the code above:

            ``@require(has_read('Example'))``
                The @require() decorator sets the permissions plugin methods.  In this example, it requires the user have read permission to the Example plugin.
            ``page_template = plugin_template(self, "templates/helloworld.mako")``
                This line sets our template to templates/helloworld.mako which exists in our Example plugin directory.  The plugin_template() function is pretty much the same as Mako's lookup_template() function except that it is smart enough to know when a template is inside of a plugin's zip archive.
            ``return self.standard_render(...``
                Every subclass of Plugin gets standard_render() for free:  It is similar to Mako's render() function except that it automatically takes care of the following:
                 - It injects the _() gettext function into the template's namespace using the proper language settings.
                 - It injects some standard PyCI variables (plugin_name, plugin_object, and language) into the template namespace.
                 - All keyword arguments passed to standard_render() other than 'template' will get added to your template's namespace in a dictionary named, "plugin_dict" (more on this later).
            Lastly, 'help', 'long_name', and 'description' are special variables used by the plugin_base.mako template that your template will likely inherit from.  Try changing them to see where they're changed on the page.
        """
        page_template = plugin_template(self, "templates/helloworld.mako")

        help=_("<p>This is the example plugin's 'Hello, world!' page.</p>")

        return self.standard_render(
            template=page_template,
            help=help,
            long_name=_('Example Plugin: Hello, world!'),
            description=_("An example plugin for developers (Hello, world!)."),
        )

    @require(has_read('Example'))
    def jqGridExample(self):
        """This example plugin page demonstrates how to use jqGrid."""
        page_template = plugin_template(self, "templates/jqgrid_example.mako")

        return self.standard_render(
            template=page_template,
            help=_('<p>This plugin page shows a basic jqGrid.</p>'),
            long_name=_('Basic jqGrid example.'),
            description=_("This is an example demonstrating how to use the jqGrid functions."),
        )

    # This method is called by jqGrid to grab the data used to fill the grid:
    @require(has_read('Example'))
    def jqgrid_example_json(self, rows=None, sidx=None, _search=None, searchField=None, searchOper=None, searchString=None, page=None, sord=None, nd=None):
        """Returns /etc/hosts in a json format that's compatible with jqGrid."""
        # This will print out (on the terminal you ran PyCI) all the arguments passed to this method
        print cherrypy.request.params # (useful for debugging)
        # Don't forget to remove the above line when moving your code into production.
        header = ["ip", "hostnames"] # This must match the 'colModel' in our jqGrid parameters (in the JavaScript)
        # The jqgrid_json() function maps directly to the parameters passed by jqGrid.  The only thing you have to
        # change when using it is the first two arguments which accept a list of lists (list_etc_hosts() in this case)
        # and a header which is used in sorting.  All other parameters should just mirror the arguments passed to the method.
        return jqgrid_json(self, list_etc_hosts(), header, rows=rows, sidx=sidx, _search=_search, searchField=searchField, searchOper=searchOper, searchString=searchString, page=page, sord=sord)
        # Note: For debugging, sometimes it is useful to do this instead of just returning the result of jqgrid_json:
        # result = jqgrid_json(self, list_etc_hosts(), header, rows=rows, sidx=sidx, _search=_search, searchField=searchField, searchOper=searchOper, searchString=searchString, page=page, sord=sord)
        # print result
        # return result

    @require(has_read('Example'))
    def NotificationExample(self):
        """This page demonstrates how to display useful messages within PyCI."""
        page_template = plugin_template(self, "templates/notification_example.mako")

        return self.standard_render(
            template=page_template,
            help=_('<p>This plugin page demonstrates how to display useful messages to the user.</p>'),
            long_name=_('Basic notification message example.'),
            description=_("Type something other than an integer into the form below then click somewhere else on the page (to un-focus the input).  Then try submitting the form with and without a valid integer value to see the notifications."),
        )

    @require(has_read('Example'))
    def validate_integer(self, number):
        """Checks whether or not the passed parameters are integers."""
        import json
        # Note: For this example I'm iterating over all passed keyword arguments even though
        #       I know we'll only be getting one (number).  Just to show how it's done.
        try:
            int(number)
            return "" # Returning nothing means this is valid
        except ValueError:
            # The Validation plugin will display this message next to the input element:
            return _('"%s is not an integer"') % `number` # Note how I wrapped the text in quotes--The validation plugin needs it

    # Note that here we're requiring the user have write permission:
    @require(has_write('Example'))
    def notification_test(self, number):
        """
        Returns a success message if 'number' is an integer, an error message if not.

        This method is used by the NotificationExample() page to display the result from the form submission
        (by way of an AJAX call using the jQuery Form plugin).
        """
        try:
            int(number)
            # Note how I wrapped each message below in the _() function (want to make sure they get translated too)
            return jquery_ui_highlight_msg(_('%s is an integer!') % number)
        except ValueError as e:
            return jquery_ui_error_msg(_('%s is not an integer!<br />%s') % (number, e))

    @require(has_read('Example'))
    def index(self):
        """
        The index() method in any Plugin is called when the user navigates directly
        to the plugin instead of one of the plugin's sub-methods.  PyCI doesn't have any
        navigation elements that would lead a user here but it is a good idea in case
        the user types in the URL directly like so:  http://192.168.1.20/Example/

        In this example plugin, index() just calls self.HelloWorld().
        """
        # Just return the HelloWorld() method:
        return self.HelloWorld()