# -*- coding: utf-8 -*-

__license__ = "GNU General Public License (GPL) Version 3"
__author__ = 'Dan McDougall <YouKnowWho@YouKnowWhat.com>'

import os, re, time
from plugin_utils import get_translation
from configfile import ConfigFile

# Take care of internationalization
trans = get_translation(__file__)
_ = lambda s: trans.ugettext(s)

def crypt_password(password, magic='$1$'):
    """Returns a shadow-valid encrypted password hash (using '$1$' magic by default)."""
    import string
    from crypt import crypt
    from random import Random
    salt = ''.join( Random().sample(string.letters+string.digits+'./', 8))
    salt = magic + salt + '$'
    hashed_pass = crypt(password, salt)
    return hashed_pass

class ShadowEntry(object):
    """Data structure to store a user's parameters in /etc/shadow.

    Here's a description of the values (taken directly from 'man shadow'):
        password_last_changed # Today, in days since Jan 1, 1970 that password was last changed
        password_days_before_change # days before password may be changed
        password_days_must_change # days after which password must be changed
        password_days_expire_warn # days before password is to expire that user is warned
        password_days_expiration_disable # days after password expires that account is disabled
        password_days_expiration # days since Jan 1, 1970 that account is disabled
    """
    def __init__(self, userobj,
        password = '!',
        # For some reason this doesn't work on an actual Routerstation:
        #password_last_changed = str(int(time.mktime(time.gmtime()) / 86400)),
        password_last_changed = str(time.mktime(time.gmtime()) / 86400).split('.')[0],
        password_days_before_change = '0',
        password_days_must_change = '99999',
        password_days_expire_warn = '7',
        password_days_expiration_disable = '',
        password_days_expiration = '',
        reserved = '' # This isn't used by anything that I know of but it's there if people want it.
    ):
        """Sets the requisite shadow values.  Must be given a valid User() object, 'userobj'."""
        self.userobj = userobj # link back to the User() object
        self.password = password # Actual password hash or a disabled marker (e.g. '!', '*')
        self.password_last_changed = password_last_changed # Today, in days since Jan 1, 1970 that password was last changed
        self.password_days_before_change = password_days_before_change # days before password may be changed
        self.password_days_must_change = password_days_must_change # days after which password must be changed
        self.password_days_expire_warn = password_days_expire_warn # days before password is to expire that user is warned
        self.password_days_expiration_disable = password_days_expiration_disable # days after password expires that account is disabled
        self.password_days_expiration = password_days_expiration # days since Jan 1, 1970 that account is disabled
        self.reserved = reserved # a reserved field

    def asList(self):
        """Returns our parameters as a list"""
        shadow_list = [
            self.userobj.username,
            self.password,
            self.password_last_changed,
            self.password_days_before_change,
            self.password_days_must_change,
            self.password_days_expire_warn,
            self.password_days_expiration_disable,
            self.password_days_expiration,
            self.reserved
        ]
        return shadow_list

    def __str__(self):
        """Returns the user's shadow line as it is in /etc/shadow"""
        return ":".join(self.asList())

    def __repr__(self):
        """Returns the user's shadow line as a string representation of a list"""
        return str(self.asList())

    def __iter__(self):
        return self.asList().__iter__()

    def __len__(self):
        return len(self.asList())

    def __contains__(self, val):
        for item in self.asList():
            if item == val:
                return True
        else:
            return False

    def __getitem__(self, val):
        """This lets you access ShadowEntry objects as a dictionary or a list.

        Example:
            passwd = PasswdFile()
            testuser = passwd.user('testuser')
            print testuser.shadow['password']
            $1$qTMPYAlL$3kX33B4cFekaRw4p7Nkpi.
            print testuser.shadow[1]
            $1$qTMPYAlL$3kX33B4cFekaRw4p7Nkpi."""
        try:
            return self.__dict__[val]
        except:
            try:
                return self.asList()[val]
            except:
                return None

class User(object):
    """Class for referencing and saving user properties (UID, GID, password, shell, etc)"""
    salt_regex = re.compile(r'.*\$.\$.*\$')
    locked_regex = re.compile(r'^\*.*')
    # This matches common strings indicating a password as disabled:
    disabled_regex = re.compile(r'([*]$)|([!][!]?$)|(x$)|([*][Ll][Kk][*])') # '*', '!', '!!', 'x', or '*LK*' (case insensitive)

    class UserException(Exception):
        """A custom exception for reporting problems performing operations within User()"""
        def __init__(self, value):
            super(UserException, self).__init__()
            self.parameter = value
        def __str__(self):
            return repr(self.parameter)

    def construct_shadow_entry(self):
        """Creates and assigns a NEW ShadowEntry() object as self.shadow using default values.

        Note:  This is primarily for when you create a new User() object from scratch."""
        if os.path.exists('/etc/shadow'):
            self.shadow = ShadowEntry(self)
        else:
            raise self.UserException(_("Shadow file is not present (shadow passwords are disabled)."))

    def get_shadow_entry(self):
        """Creates a new ShadowEntry() object from the user's parameters in /etc/shadow and assigns it to self.shadowobj.

        Note: This should only be used when creating a new user from scratch outside of a PasswdFile() object."""
        if not os.path.exists('/etc/shadow'):
            raise self.UserException(_("Shadow file is not present (shadow passwords are disabled)."))

        for line in open('/etc/shadow', 'r').readlines():
            line = line.strip('\n').split(':')
            if line[0] == self.username:
                password = line[1]
                password_last_changed = line[2]
                password_days_before_change = line[3]
                password_days_must_change = line[4]
                password_days_expire_warn = line[5]
                password_days_expiration_disable = line[6]
                password_days_expiration = line[7]
                reserved = line[8]
        self.shadow = ShadowEntry(self,
            password,
            password_last_changed,
            password_days_before_change,
            password_days_must_change,
            password_days_expire_warn,
            password_days_expiration_disable,
            password_days_expiration,
            reserved
        )

    def get_password_status(self):
        """Sets self.password_status based on the password of this User object (dealing with /etc/shadow if necessary)"""
        if self.shadow:
            if not self.shadow.password or self.disabled_regex.match(self.shadow.password):
                self.password_status = _("Disabled")
            elif self.shadow.password == '':
                self.password_status = _("Empty Password")
            elif self.salt_regex.match(self.shadow.password): # This password is hashed
                if self.shadow.password.startswith('*'): # If this contains a valid hash and it starts with * that means the account is locked
                    self.password_status = _("Locked")
                else:
                    self.password_status = _("Password")
            else:
                self.password_status = _("Cleartext") # I'm leaving it up to the programmer to decide when to hash it
        else:
            if not self.password or self.disabled_regex.match(self.password):
                self.password_status = _("Disabled")
            elif self.password == '':
                self.password_status = _("Empty Password")
            elif self.salt_regex.match(self.password): # This password is hashed
                if self.password.startswith('*'): # If this contains a valid hash and it starts with * that means the account is locked
                    self.password_status = _("Locked")
                else:
                    self.password_status = _("Password")
            else:
                self.password_status = _("Cleartext") # I'm leaving it up to the programmer to decide when to hash it

    def __init__(self, username, password, uid, gid, gecos, homedir, shell):
        """Set up a bunch of user properties and link them to a passwd/shadow file objects"""
        self.username = username
        self.password = password # This will just be 'x' if shadow is enabled
        self.uid = uid
        self.gid = gid
        self.gecos = gecos
        self.homedir = homedir
        self.shell = shell
        self.modified = False
        self.shadow = None
        self.get_password_status()

    def __str__(self):
        return ":".join(self.asList())

    def __repr__(self):
        return str(self.asList())

    def __iter__(self):
        return self.asList().__iter__()

    def __len__(self):
        return len(self.asList())

    def __contains__(self, val):
        for item in self.asList():
            if item == val:
                return True
        else:
            return False

    def __getitem__(self, val):
        """This lets you access User objects as a dictionary or a list.

        Example:
            passwd = PasswdFile()
            testuser = passwd.user('testuser')
            print testuser['shell']
            /bin/bash
            print testuser[6]
            /bin/bash"""
        try:
            return self.__dict__[val]
        except:
            try:
                return self.asList()[val]
            except:
                return None

    def asList(self):
        """Returns the user's parameters as a list.

        Note:  Does not include self.shadow."""
        return [self.username, self.password, self.uid, self.gid, self.gecos, self.homedir, self.shell]

    def hash_password(self, magic='$1$'):
        """Hash the user's password if it isn't already.  Stores it as self.password or self.shadow.password appropriately."""
        if os.path.exists('/etc/shadow'):
            hashed_password = crypt_password(self.shadow.password, magic)
            self.shadow.password = hashed_password
        else:
            hashed_password = crypt_password(self.password, magic)
            self.password = hashed_password
        self.password_status = _("Password")

    def save(self):
        """Saves/updates the user's properties in /etc/passwd and /etc/shadow"""
        passwd = PasswdFile()
        updating = False # This is used when we are making an update to an existing record (as opposed to adding a new one)
        for index, userobj in enumerate(passwd.asList):
            if userobj.username == self.username:
                passwd.asList[index] = self
                updating = True
                break
        if updating: # We're done
            passwd.save()
        else: # We're appending
            passwd.asList.append(self)
            passwd.save()

    def lock(self):
        """Locks the user's account by placing an asterisk (*) at the beginning of their password hash.

        Example:  '$1$6zC7IDVb$D0KADmAQwgy4thyPjrwn31' becomes '*$1$6zC7IDVb$D0KADmAQwgy4thyPjrwn31'

        This prevents the user from being able to use a password to login."""
        if os.path.exists('/etc/shadow'):
            if not self.shadow.password.startswith('*'):
                self.shadow.password = '*' + self.shadow.password
                self.password_status = _("Locked")
            else:
                raise self.UserException(_("User is already locked."))
        else:
            if not self.password.startswith('*'):
                self.password = '*' + self.password
                self.password_status = _("Locked")
            else:
                raise self.UserException(_("User is already locked."))

    def unlock(self):
        """Unlocks the user's account by removing the leading asterisk in their password hash."""
        if os.path.exists('/etc/shadow'):
            if self.shadow.password.startswith('*'):
                self.shadow.password = self.shadow.password[1:]
                self.password_status = _("Password")
            else:
                raise self.UserException(_("User is not locked."))
        else:
            if self.password.startswith('*'):
                self.password = self.password[1:]
                self.password_status = _("Password")
            else:
                raise self.UserException(_("User is not locked."))

    def disable(self):
        """Permanently disables the user's account by replacing their password hash with '!'"""
        if os.path.exists('/etc/shadow'):
            self.shadow.password = "!"
            self.password_status = _("Disabled")
        else:
            self.password = "!"
            self.password_status = _("Disabled")

    def delete(self):
        """Permanently delete the user from /etc/passwd and /etc/shadow (if enabled)"""
        passwd = PasswdFile()
        for userobj in passwd:
            if userobj.username == self.username:
                passwd.remove(userobj)
                break
        passwd.save()
# TODO:  Add a groups() function to User()

    def expire(self, days=None):
        """Sets a user's password expiration date (in days since Jan 1 1970).
        If no 'days' variable is given it immediately expires their password.
        Note:  Only works on systems with /etc/shadow."""
        if not days:
            self.shadow.password_expiration = str(int(time.mktime(time.gmtime()) / 86400)) # Equivalent to 'usermod --expiredate 1'
        else:
            self.shadow.password_expiration = days

# Classes
class PasswdFile(ConfigFile):
    """ConfigFile class to read and write to /etc/passwd and /etc/shadow"""

    def read_passwd(self):
        """Read /etc/passwd and return it as a list of User objects."""
        out_list = []
        for line in open(self.filepath, 'r').readlines():
            fields = line.rstrip('\n').split(':')
            userobj = User(fields[0], fields[1], fields[2], fields[3], fields[4], fields[5], fields[6])
            out_list.append(userobj)
        return out_list

    def read_shadow(self):
        """Read /etc/shadow and assign each User() a ShadowEntry() object based on the found information.
        Also returns /etc/shadow as a list (useful for doing comparisons).

        Note:  We don't use User.get_shadow_entry() to cut down on filesystem reads."""
        shadow = []
        for line in open('/etc/shadow', 'r').readlines():
            fields = line.rstrip('\n').split(':')
            for userobj in self.asList:
                if userobj.username == fields[0]:
                    userobj.shadow = ShadowEntry(userobj,
                        fields[1],
                        fields[2],
                        fields[3],
                        fields[4],
                        fields[5],
                        fields[6],
                        fields[7]
                    )
                    userobj.get_password_status()
                    break
            shadow.append(fields)
        return shadow

    def __init__(self):
        """Sets up some basic properties and executes self.read_passwd()."""
        self.filepath = '/etc/passwd'
        self.asList = self.read_passwd()
        if os.path.exists('/etc/shadow'):
            # Read and assign user's their shadow parameters
            self.shadow_enabled = True
            self.shadow = self.read_shadow()
        else:
            self.shadow_enabled = False

    def __str__(self):
        out_string = ""
        for userobj in self.asList:
            out_string += str(userobj) + '\n'
        return out_string.rstrip('\n')

    def __repr__(self):
        return str(self.asList)

    def __iter__(self):
        return self.asList.__iter__()

    def __len__(self):
        return len(self.asList)

    def __contains__(self, user):
        """This let's you do things like this:
            if 'someuser' in passwd:
                print "someuser is present in passwd"
        ...and things like this:
            someuser = User('someuser', 'x', '1111', '100', '', '/home/someuser', '/bin/bash')
            if someuser in passwd:
                print "someuser is present in passwd"

        Note: This obviously is only designed to work with usernames
        """
        if isinstance(user, str):
            for userobj in self.asList:
                if userobj.username == user:
                    return True
        elif isinstance(user, User):
            for userobj in self.asList:
                if userobj.username == user.username:
                    return True
        else:
            return False

    def __getitem__(self, user):
        """This lets you access PasswdFile as a dictionary (of users).

        Example:
            passwd = PasswdFile()
            print passwd['root']
            root:$1$uHBYX2c0$1DxS5C7iB36HIXA.DcH2p0:0:0:root:/root:/bin/bash"""
        if isinstance(user, str):
            for userobj in self.asList:
                if userobj.username == user:
                    return userobj
        elif isinstance(user, User):
            for userobj in self.asList:
                if userobj.username == user.username:
                    return userobj

    def __setitem__(self, key, user):
        """This lets you do item (User()) assignment in PasswdFile objects.

        Example:
            passwd = PasswdFile()
            testuser = passwd.user('testuser')
            testuser.gecos = "testing"
            new_passwd = PasswdFile()
            new_passwd['testuser'] = testuser
            new_passwd.save()"""
        # Note:  For some reason I can't figure out another way to do this.  'self[key] = user' results in infinite recursion
        for index, userobj in enumerate(self.asList):
            if userobj.username == user.username:
                self.asList[index] = user

    def remove(self, userobj):
        """Meant to emulate the remove() built-in (for lists)."""
        self.asList.remove(userobj)

    def users(self):
        """Returns a list of users in /etc/passwd"""
        out_list = []
        for userline in self.asList:
            out_list.append(userline[0])
        return out_list

    def uids(self):
        """Returns a list of uids in /etc/passwd"""
        out_list = []
        for userline in self.asList:
            out_list.append(userline[2])
        return out_list

    def gids(self):
        """Returns a list of gids in /etc/passwd"""
        out_list = []
        for userline in self.asList:
            out_list.append(userline[3])
        return out_list

    def comments(self):
        """Returns a list of comments in /etc/passwd"""
        out_list = []
        for userline in self.asList:
            if len(userline[4].strip()) > 0: # Only add it to the list if it isn't empty
                out_list.append(userline[4])
        return out_list

    def homedirs(self):
        """Returns a list of unique home directories in /etc/passwd"""
        out_list = []
        for userline in self.asList:
            out_list.append(userline[5])
        return list(set(out_list))

    def shells(self):
        """Returns a list of unique shells in /etc/passwd"""
        out_list = []
        for userline in self.asList:
            out_list.append(userline[6])
        return list(set(out_list))

    def user(self, username):
        """If the user exists, returns a User object generated from said user's properties in /etc/passwd."""
        for userobj in self.asList:
            if userobj.username == username:
                return userobj

    def save(self):
        """Overwrites /etc/passwd with the data from self.asList().  Also stores any hashed passwords in /etc/shadow if present."""
        passwd_ondisk = PasswdFile() # Create a copy of ourselves from the on-disk files
        if str(self) == str(passwd_ondisk): # Check for changes from /etc/passwd
            if not self.shadow_enabled:
                return _("No changes to save in /etc/passwd")
            else: # Now check for changes to user's shadow parameters
                shadow = open('/etc/shadow','r').read()
                changes = False
                for userobj in self.asList:
                    if str(userobj.shadow) not in shadow:
                        changes = True
                        break
                if not changes:
                    return _("No changes to save in /etc/passwd or /etc/shadow")
        if not self.shadow_enabled:
            self.save_delimited(self.asList, '/etc/passwd', ':')
            return _("Saved passwd successfully")
        else:
            # First save the passwd file
            self.save_delimited(self.asList, '/etc/passwd', ':')
            # Now save the shadow file
            new_shadow = []
            for userobj in self.asList:
                new_shadow.append(userobj.shadow.asList())
            self.save_delimited(new_shadow, '/etc/shadow', ':')
        self.asList = self.read_passwd()

def list_etc_passwd():
    """Returns /etc/passwd as a list of strings."""
    passwd = PasswdFile()
    user_list = []
    for userobj in passwd.asList:
        username = userobj.username
        password_status = get_user_password_status(userobj[0]) # Replace password placeholder with the password status
        uid = int(userobj[2]) # Convert UID from string to int so sorting works properly
        gid = int(userobj[3]) # Ditto for GID
        gecos = userobj[4]
        homedir = userobj[5]
        shell = userobj[6]
        user_list.append([username, password_status, uid, gid, gecos, homedir, shell])
    return user_list

# TODO:  Figure out how to determine if a user is locked after too many failed login attempts
def get_user_password_status(username):
    """Returns the password status ('Password', 'Empty Password', or 'Disabled') of the given user."""
    passwd = PasswdFile()
    return passwd.user(username).password_status