#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Description: Preferences plugin.  For setting user's preferences.

import plugin_utils
import cherrypy
import os
from auth import require, has_read, has_write
from pyci_db import connect_db, disconnect_db
from utils import lookup_template, search_cells, jqgrid_format, jqgrid_select_options, plugin_template, jquery_ui_error_msg, jquery_ui_highlight_msg, format_columns

# Internationalization
trans = plugin_utils.get_translation('preferences')
_ = lambda s: trans.ugettext(s)

# TODO:  Figure out a way to allow users to specify language preferences as opposed to globally setting them.
# Note:  I abandoned per-user language preferences for the time being until I figure out the TODO above.
def verify_preferences():
    """Verifies that the user_preferences table exists in the PyCI DB and if not, creates it."""
    print _("Verifying user preferences table...")
    pyci_db = connect_db('pyci.db')
    dbcursor = pyci_db.cursor()
    # These silently set up the tables if they don't already exist
    # Create tables...
    dbcursor.execute('CREATE TABLE IF NOT EXISTS user_preferences (id INTEGER PRIMARY KEY, user TEXT, plugin TEXT, preference TEXT, setting TEXT)')
    # Create indexes to keep things speedy...
    dbcursor.execute('CREATE INDEX IF NOT EXISTS user_preferences_index ON user_preferences (preference)')
    dbcursor.execute('CREATE INDEX IF NOT EXISTS user_preferences_users_index ON user_preferences (user)')
    dbcursor.execute('CREATE INDEX IF NOT EXISTS user_preferences_plugin_index ON user_preferences (plugin)')
    dbcursor.execute('CREATE INDEX IF NOT EXISTS user_preferences_preference_index ON user_preferences (preference)')
    disconnect_db(pyci_db)

def get_language_preference(user):
    """Returns the language preference of the current logged-in user.  If no language preference is set, return None"""
    pyci_db = connect_db('pyci.db')
    dbcursor = pyci_db.cursor()
    try: # This will raise an exception if there's no record
        record = dbcursor.execute('SELECT setting FROM user_preferences WHERE user=(?) AND preference="preferred_language"', (user,)).fetchone()[0]
        disconnect_db(pyci_db)
        return record
    except:
        return None

def set_language_preference(user, language, id=None):
    """Sets a user's language preference."""
    pyci_db = connect_db('pyci.db')
    dbcursor = pyci_db.cursor()
    try:
        if id:
            dbcursor.execute("UPDATE user_preferences SET setting=(?) WHERE id=(?)", (language, id))
        else:
            dbcursor.execute("INSERT INTO user_preferences VALUES (null, ?, 'Preferences', 'preferred_language', ?)", (user, language))
        disconnect_db(pyci_db)
        return True
    except:
        return False

class Preferences(plugin_utils.Plugin):
    """A plugin for OpenWRT's network configuration"""
    #permissions = ['Preferences',]
    display = False
    static = 'plugins/preferences/static'
    #sort_priority = 1
    plugin_dict = {
        'long_name': _('User Preferences'),
        'description': _('User Preferences:  This is where you can set global PyCI preferences for your user account such as your preferred language.'),
    }
    navigable_functions = {_('Preferred Language'): 'Language'}

    def __init__(self):
        """Verify the preferences table exists in the PyCI database."""
        verify_preferences()
        # Note:  This is a good example of how to store your plugin settings and such in the DB.

    @require() # We don't require any particular permission since these settings apply only to the logged-in user
    def Language(self, **data):
        """Render the preferences configuration page"""
        language_template = plugin_template(self, "templates/language.mako")
        valid_languages = []
        for item in os.listdir('./i18n'):
            if os.path.isdir('i18n/%s' % item) and not item.startswith("."):
                valid_languages.append(item)

        return self.standard_render(
            template=language_template,
            valid_languages=valid_languages,
            preferred_language=get_language_preference(cherrypy.request.login),
            help=_('Preferred Language help.  TODO:  Write this.'),
            long_name=_('Preferred Language'),
            description=_("Select your preferred language so that PyCI may use it if available.  Only available languages will be listed."),
        )

    @require()
    def set_language(self, language):
        """Saves the user's language preference in the PyCI DB"""
        user = cherrypy.request.login
        pyci_db = connect_db('pyci.db')
        dbcursor = pyci_db.cursor()
        try:
            # Get the existing record ID if one already exists
            record = dbcursor.execute('SELECT id FROM user_preferences WHERE user=(?) AND preference="preferred_language"', (user,)).fetchone()
            disconnect_db(pyci_db)
            if record:
                set_language_preference(user=user, language=language, id=record[0])
            set_language_preference(user, language)
            os.environ['LANG'] = language
            return jquery_ui_highlight_msg(_('<strong>Success:</strong> Language preference has been updated.'))
        except Exception as e:
            print e
            return jquery_ui_error_msg(_('Could not save your language preference.  Please make sure PyCI is running as root.<br />%s' % e))

    @require()
    def index(self):
        """Return the index page"""
        return self.Language()
