# -*- coding: utf-8 -*-

__license__ = "GNU General Public License (GPL) Version 3"
__author__ = 'Dan McDougall <YouKnowWho@YouKnowWhat.com>'

import os, sys, gettext
from time import sleep
from datetime import datetime, timedelta
import calendar
try: # Note: The json module is only present in Python 2.6+
    import json
except: # If not found, try importing simplejson which will be required if using Python 2.4 or Python 2.5
    import simplejson as json
# 3rd party modules
import cherrypy
# PyCI modules
from plugin_utils import get_translation, Plugin
from auth import require, has_read, has_write
from utils import lookup_template, plugin_template, jquery_ui_error_msg, jquery_ui_highlight_msg, sortby_field, split_seq, jqgrid_json
from iptables import IPTables
from odict import odict

"""Status Plugin -- Get the status of various items and operations on your host."""

# Take care of internationalization
trans = get_translation(__file__)
_ = lambda s: trans.ugettext(s)

def cpu_utilization(cpu='cpu ', interval=1):
    """Returns a dict representing each respective field for 'cpu' in /proc/stat.
    Must run twice to return valid results.
    Note: To get results on all CPUs we need to pass 'cpu ' (note the space).
          This is the default.  For other CPUs you can specify the number, 'cpu1'
    @cpu:  The start of the line in /proc/stat that corresponds with the cpu you want to monitor
    @interval:  That amount of time to wait between checks to calculate the utilization."""
    cpu_line = [line for line in open('/proc/stat') if line.startswith(cpu)][0]
    cpu_user1 = float(cpu_line.split()[1])
    cpu_nice1 = float(cpu_line.split()[2])
    cpu_sys1 = float(cpu_line.split()[3])
    cpu_idle1 = float(cpu_line.split()[4])
    cpu_iowait1 = float(cpu_line.split()[5])
    cpu_irq1 = float(cpu_line.split()[6])
    cpu_softirq1 = float(cpu_line.split()[7])
    cpu_total1 = sum(map(float, cpu_line.split()[1:8]))
    sleep(interval)

    cpu_line = [line for line in open('/proc/stat') if line.startswith(cpu)][0]
    cpu_user2 = float(cpu_line.split()[1])
    cpu_nice2 = float(cpu_line.split()[2])
    cpu_sys2 = float(cpu_line.split()[3])
    cpu_idle2 = float(cpu_line.split()[4])
    cpu_iowait2 = float(cpu_line.split()[5])
    cpu_irq2 = float(cpu_line.split()[6])
    cpu_softirq2 = float(cpu_line.split()[7])
    cpu_total2 = sum(map(float, cpu_line.split()[1:8]))

    diff_user = cpu_user1 - cpu_user2
    diff_nice = cpu_nice1 - cpu_nice2
    diff_sys = cpu_sys1 - cpu_sys2
    diff_idle = cpu_idle1 - cpu_idle2
    diff_iowait = cpu_iowait1 - cpu_iowait2
    diff_irq = cpu_irq1 - cpu_irq2
    diff_softirq = cpu_softirq1 - cpu_softirq2
    diff_total = cpu_total1 - cpu_total2

    out_dict = {}
    out_dict['cpu_user'] = (1000 * diff_user / diff_total + 5)/10
    out_dict['cpu_nice'] = (1000 * diff_nice / diff_total + 5)/10
    out_dict['cpu_sys'] = (1000 * diff_sys / diff_total + 5)/10
    out_dict['cpu_idle'] = (1000 * diff_idle / diff_total + 5)/10
    out_dict['cpu_iowait'] = (1000 * diff_iowait / diff_total + 5)/10
    out_dict['cpu_irq'] = (1000 * diff_irq / diff_total + 5)/10
    out_dict['cpu_softirq'] = (1000 * diff_softirq / diff_total + 5)/10
    out_dict['cpu_usage'] = (
        out_dict['cpu_user'] +
        out_dict['cpu_nice'] +
        out_dict['cpu_sys'] +
        out_dict['cpu_iowait'] +
        out_dict['cpu_irq'] +
        out_dict['cpu_softirq']
        )/1
    return out_dict

def getloadavg():
    """Emulates os.getloadavg() since the version of Python on OpenWRT doesn't have it.
    ...which is bizarre."""
    return open('/proc/loadavg', 'r').read().split()[:3]

def get_bandwidth():
    """Grabs packet data from /proc/net/dev and returns it as a dict"""
    netdev = open('/proc/net/dev', 'r')
    interface_dict = {}
    for line in netdev:
        if not line.startswith('Inter-') and not line.startswith(' face'): # Don't want the headers
            line = line.split(':')
            interface_dict[line[0].strip()] = {
                'bytes_received': int(line[1].split()[0]),
                'bytes_sent': int(line[1].split()[8])}
    return interface_dict

def memory_status():
    """Returns a dict representing the status of memory allocation on this system"""
    proc_meminfo = open('/proc/meminfo', 'r')
    meminfo_dict = {}
    for line in proc_meminfo:
        line = line.split()
        meminfo_dict.update({line[0].rstrip(':'): line[1]})
    return meminfo_dict

def list_chain(iptables_obj, table_name, chain_name, sortname=None, sortorder=None):
    """
    Returns the given chain from the given table in list format.
    Sorted by 'sortname' in 'sortorder' order (either 'desc' or 'asc').

    Note:  'table_name' should be one of 'filter', 'nat', 'mangle', or 'raw'.
    """
    chain_list = []
    for ruleobj in iptables_obj.tables[table_name][chain_name]:
        chain_list.append([
            ruleobj.ruleno,
            ruleobj.packets,
            ruleobj.bytes,
            ruleobj.target,
            ruleobj.protocol,
            ruleobj.target_opts,
            ruleobj.in_interface,
            ruleobj.out_interface,
            ruleobj.source,
            ruleobj.destination,
            ruleobj.status_params
        ])
    if sortname:
        fields = ['num', 'pkts', 'bytes', 'target', 'prot', 'opt', 'in', 'out',
                  'source', 'destination', 'status_params']
        chain_list = sortby_field(sortname, fields, chain_list)
    if sortorder == "desc":
        chain_list.reverse()
    return chain_list

def get_netstat():
    """Returns the result of 'netstat -pln' as a list of lists
    (suitable for use in jqGrid)
    Each list (in the list) uses the following format:
    ['Proto', 'Recv-Q', 'Send-Q', 'Local Address', 'Foreign Address', 'State', 'PID/Program name']"""
    from commands import getoutput
    out_list = []
    for line in getoutput('netstat -pln').splitlines():
        value_list = []
        if line:
            if line.startswith('tcp') or line.startswith('udp'):
                out_list.append(line.split())
    return out_list

class Status(Plugin):
    """A plugin for geting the status of various items and operations on your host."""
    display = True
    permissions = ['Status',]
    sort_priority = 1
    navigable_functions = odict([
        (_('Firewall'), 'Firewall'),
        (_('System Monitors'), 'SystemMonitors'),
        (_('Listening Services'), 'ListeningServices'),
    ])

    previous_bandwidth = None

    last_run = datetime.now() # This is used to breifly cache the IPTables() object for use in iptables_json
    iptables = None # Ditto

    @require(has_read(permissions[0]))
    def SystemMonitors(self, **data):
        """Render the system monitors page"""
        _template = plugin_template(self, "templates/monitors.mako")
        meminfo_dict = memory_status()

        return self.standard_render(
            template=_template,
            help=_("This page represents various monitoring points on this host."),
            long_name=_('Operating System Status'),
            description=_("Various charts and graphs representing various operating system data points."),
            meminfo_dict=meminfo_dict,
        )

    @require(has_read(permissions[0]))
    def load_average(self, **options):
        """Return the current system load average in a json format flot understands."""
        now = calendar.timegm(datetime.now().timetuple()) * 1000
        json_out = []
        json_out.append({'data':[[now, getloadavg()[0]]], 'label': '5 minute'})
        json_out.append({'data':[[now, getloadavg()[1]]], 'label': '10 minute'})
        json_out.append({'data':[[now, getloadavg()[2]]], 'label': '15 minute'})
        return json.dumps(json_out)

    @require(has_read(permissions[0]))
    def cpu_load(self, **options):
        """Return the current CPU utilization stats in a json format flot understands."""
        now = calendar.timegm(datetime.now().timetuple()) * 1000
        json_out = []
        load_dict = cpu_utilization()
        json_out.append({'data':[[now, load_dict['cpu_usage']]], 'label': '%CPU'})
        json_out.append({'data':[[now, load_dict['cpu_sys']]], 'label': '%sys'})
        json_out.append({'data':[[now, load_dict['cpu_user']]], 'label': '%user'})
        json_out.append({'data':[[now, load_dict['cpu_idle']]], 'label': '%idle'})
        json_out.append({'data':[[now, load_dict['cpu_iowait']]], 'label': '%iowait'})
        json_out.append({'data':[[now, load_dict['cpu_nice']]], 'label': '%nice'})
        json_out.append({'data':[[now, load_dict['cpu_irq']]], 'label': '%irq'})
        json_out.append({'data':[[now, load_dict['cpu_softirq']]], 'label': '%softirq'})
        return json.dumps(json_out)

    @require(has_read(permissions[0]))
    def bandwidth(self, **options):
        """Return the current system load average in a json format flot understands.
        Note:  Only returns interfaces with counters greater than 0."""
        now = calendar.timegm(datetime.now().timetuple()) * 1000
        json_out = []
        bandwidth_now = get_bandwidth()
        bandwidth_difference = {}
        if self.previous_bandwidth: # Compare the previous run with this one so we can calculate the difference
            for key in self.previous_bandwidth.keys():
                bandwidth_difference[key] = {}
                bandwidth_difference[key]['bytes_received'] = bandwidth_now[key]['bytes_received'] - self.previous_bandwidth[key]['bytes_received']
                bandwidth_difference[key]['bytes_sent'] = bandwidth_now[key]['bytes_sent'] - self.previous_bandwidth[key]['bytes_sent']
        else:
            for key in bandwidth_now.keys():
                bandwidth_difference[key] = {}
                bandwidth_difference[key]['bytes_received'] = 0
                bandwidth_difference[key]['bytes_sent'] = 0
        self.previous_bandwidth = bandwidth_now
        for key in bandwidth_difference.keys():
            in_label = '%s in' % key
            out_label = '%s out' % key
            if bandwidth_now[key]['bytes_received'] + bandwidth_now[key]['bytes_sent'] > 0:
                json_out.append({'data':[[now, bandwidth_difference[key]['bytes_received']]], 'label': in_label})
                json_out.append({'data':[[now, bandwidth_difference[key]['bytes_sent']]], 'label': out_label})
        return json.dumps(json_out)

    @require(has_read(permissions[0]))
    def meminfo(self, totalmem=None, **options):
        """Return the current memory utilization in a json format flot understands."""
        now = calendar.timegm(datetime.now().timetuple()) * 1000
        meminfo_dict = memory_status()
        json_out = []
        if totalmem:
            return json.dumps([int(meminfo_dict['MemTotal']),])
        json_out.append({'data':[[now, int(meminfo_dict['MemFree'])/1024]], 'label': 'Free MB'})
        return json.dumps(json_out)

    @require(has_read('Status'))
    def Firewall(self):
        """Render the firewall status page"""
        firewall_template = plugin_template(self, "templates/firewall.mako")
        # TODO:  Add a secondary grid below each table grid that shows details on the 'match' portion of each firewall rule (that allows inline editing)

        return self.standard_render(
            template=firewall_template,
            help=_('This is the help for the firewall configuration section.'),
            long_name=_('Firewall Status'),
            description=_("The status of the running firewall can be viewed in the grid below."),
        )

    @require(has_read('Status'))
    def iptables_json(self,
        table=None,
        chain=None,
        rows=None,
        sidx=None,
        _search=None,
        searchField=None,
        searchOper=None,
        searchString=None,
        page=None,
        sord=None,
        nd=None):
        """
        Returns the specified chain from the specified table from the host's
        iptables (firewall) rules in a json format that's compatible with jqGrid.
        """
        # This logic allows us to cache the IPTables() object breifly so we don't have to parse the iptables command every time this method is called.
        # ...and this method is typically called many times as the template iterates over every table and chain.
        # TODO:  Have it empty out this cached data after a timeout
        five_seconds = timedelta(seconds=5)
        if datetime.now() > self.last_run + five_seconds:
            self.last_run = datetime.now()
            self.iptables = IPTables()
        elif not self.iptables:
            self.iptables = IPTables()
        if not page:
            page = 1
        header = ['num', 'pkts', 'bytes', 'target', 'prot', 'opt', 'in', 'out',
                  'source', 'destination', 'status_params']
        iptables_list = list_chain(self.iptables, table, chain, sidx, sord)
        return jqgrid_json(
            self,
            iptables_list,
            header,
            rows=rows,
            sidx=sidx,
            _search=_search,
            searchField=searchField,
            searchOper=searchOper,
            searchString=searchString,
            page=page,
            sord=sord
        )

    @require(has_read('Status'))
    def filter_table_tab(self):
        """Render the (Firewall) filter table status page"""
        _template = plugin_template(self, "templates/firewall_table_tab.mako")
        iptables = IPTables()

        return self.standard_render(
            template=_template,
            iptables=iptables,
            table='filter',
        )

    @require(has_read('Status'))
    def nat_table_tab(self):
        """Render the (Firewall) nat table status page"""
        _template = plugin_template(self, "templates/firewall_table_tab.mako")
        iptables = IPTables()

        return self.standard_render(
            template=_template,
            iptables=iptables,
            table='nat',
        )

    @require(has_read('Status'))
    def raw_table_tab(self):
        """Render the (Firewall) raw table status page"""
        _template = plugin_template(self, "templates/firewall_table_tab.mako")
        iptables = IPTables()

        return self.standard_render(
            template=_template,
            iptables=iptables,
            table='raw',
        )

    @require(has_read('Status'))
    def mangle_table_tab(self):
        """Render the (Firewall) mangle table status page"""
        _template = plugin_template(self, "templates/firewall_table_tab.mako")
        iptables = IPTables()

        return self.standard_render(
            template=_template,
            iptables=iptables,
            table='mangle',
        )

    @require(has_read('Status'))
    def get_firewall_spreadsheet(self, table=None, chain=None):
        """Exports the user permissions table as a spreadsheet"""
        spreadsheet_template = lookup_template('html', 'spreadsheet.mako')
        cherrypy.response.headers['Content-Type'] = 'application/vnd.ms-excel; charset=utf-8'
        headers = ['num', 'pkts', 'bytes', 'target', 'prot', 'opt', 'in',
                   'out', 'source', 'destination', 'status_params']
        iptables = IPTables()
        rows = list_chain(iptables, table, chain)
        return spreadsheet_template.render(headers=headers, rows=rows)

    @require(has_read('Status'))
    def ListeningServices(self):
        """Render the listening services page"""
        _template = plugin_template(self, "templates/netstat.mako")

        return self.standard_render(
            template=_template,
            help=_('This is the help for the Listening Services section (netstat output).'),
            long_name=_('Listening Services'),
            description=_("The output of netstat (listening services) can be seen in the grid below."),
        )

    @require(has_read('Status'))
    def netstat_json(self, rows=None, sidx=None, _search=None, searchField=None, searchOper=None, searchString=None, page=None, sord=None, nd=None):
        """Returns the output of 'netstat -pln' in a json format that's compatible with jqGrid."""
        header = [_("Protocol"), _("Receive Queue"), _("Send Queue"), _("Local Address"), _("Remote Address"), _("State"), _("PID/Program")]
        netstat = get_netstat()
        print jqgrid_json(self, netstat, header, rows=rows, sidx=sidx, _search=_search, searchField=searchField, searchOper=searchOper, searchString=searchString, page=page, sord=sord)
        return jqgrid_json(self, netstat, header, rows=rows, sidx=sidx, _search=_search, searchField=searchField, searchOper=searchOper, searchString=searchString, page=page, sord=sord)