# -*- coding: utf-8 -*-
from subprocess import Popen, PIPE
import os, sys, json
import tarfile
import cherrypy
from mako.template import Template
from mako.lookup import TemplateLookup

__license__ = "GNU General Public License (GPL) Version 3"
__author__ = 'Dan McDougall <YouKnowWho@YouKnowWhat.com>'

# TODO:  Figure out if I still need to base64-encode objects into templates. I can probably remove these methods
def encode_object(obj):
    """Pickles and then Base64-encodes an object so it can be embedded in forms"""
    return base64.b64encode(pickle.dumps(obj))

def decode_object(string):
    """Un-pickles and then Base64-decodes an object"""
    return pickle.loads(base64.b64decode(string))

def normalizeString(string):
    """Removes/fixes leading/trailing newlines/whitespace and escapes double quotes with double quotes (to comply with CSV format)"""
    string = re.sub(r'(\r\n|\r|\n)', '\n', string) # Convert all newlines to unix newlines
    string = string.rstrip() # Remove trailing whitespace/blank lines
    return string

def split_uppercase(string):
    """Splits TitleCasedString into 'Title Cased String'"""
    x=''
    for i in string:
        if i.isupper():
            x+=' %s' % i
        else: x+=i
    return x.strip()

def get_display_name(plugin):
    """Returns the display_name value of the given plugin.  If none is set, return
    split_uppercase(plugin.__name__)."""
    try:
        return plugin.display_name
    except AttributeError:
        return split_uppercase(plugin.__name__)

def strip_pound(string):
    """Removes leading # signs from basic anchor links (for use in navigation)"""
    p = re.compile('^#')
    return p.sub('', string)

def jeditable_dict(_list, default=None):
    """Converts a given list into a format jeditable likes"""
    output_dict = {}
    for item in _list:
        output_dict.update({item:item})
    if default:
        output_dict.update({'selected': default})
    return json.dumps(output_dict)

# TODO:  Figure out a way to use paths other than "static"
def connect_tarball_routes(dispatcher, controller, tarball, action=None, plugin_name=None):
    """Sets up routes for a given tarball using the given dispatcher and assign it 'action' (i.e. the path on the webserver)"""
    tar = tarfile.open(tarball)
    if not action:
        # Setup a basic action based on the tarball filename
        (action, ext) = os.path.splitext(tarball)
    for tarobj in tar.getmembers():
        if tarobj.isdir():
            pass # Skip directories
        else:
            if plugin_name:
                static_path = "/%s/static/%s" % (plugin_name, tarobj.name) # e.g. /admin/static/filepath
            else:
                if action:
                    static_path = "/%s/%s" %(action, tarobj.name) # e.g. html/filepath
                else:
                    static_path = "/%s" % tarobj.name # e.g. /filepath
            dispatcher.connect(tarobj.name, static_path, controller=controller, action=action, filepath=tarobj.name)

def detect_mime_type(filepath):
    """Attempts to detect the mime type of the given file and sets the appropriate cherrypy response header"""
    f, extension = os.path.splitext(filepath)
    if extension == '.css': # It seems CherryPy can't automatically figure out what mime types are inside tarballs
        cherrypy.response.headers['Content-Type'] = "text/css"
    elif extension == '.js':
        cherrypy.response.headers['Content-Type'] = "text/javascript"
    elif extension == '.png':
        cherrypy.response.headers['Content-Type'] = "image/png"
    elif extension == '.gif':
        cherrypy.response.headers['Content-Type'] = "image/gif"
    elif extension == '.jpeg' or extension == '.jpeg':
        cherrypy.response.headers['Content-Type'] = "image/jpeg"

def get_tarball_file(tarball, filepath):
    """Returns files from within a gzip-compressed tar file (e.g. static.tgz).
    Only intended for returning static content such as images and text files.
    Note:  Will not work with Mako templates."""
    tar = tarfile.open(tarball)
    fileobj = tar.extractfile(filepath)
    detect_mime_type(filepath)
    filedata = fileobj.read()
    return filedata

def get_static_file(filepath):
    """Serves up static content from a direct filepath."""
    print filepath
    if os.path.exists(filepath):
        detect_mime_type(filepath)
        return open(filepath).read()
    else:
        raise cherrypy.HTTPError(404)

def serve_static(static, filepath):
    """Serves up static content either from within a tarball or a local directory depending on the detected file type of 'static'"""
    (plugin, ext) = os.path.splitext(static)
    if ext:
        return get_tarball_file(static, filepath)
    else: # This is a directory
        return get_static_file(os.path.join(static, filepath))

def lookup_template(basedir, _template):
    """Returns a mako Template from the html directory"""
    template_lookup = TemplateLookup(directories=[basedir], input_encoding='utf-8', output_encoding='utf-8')
    if os.path.exists(basedir):
        return template_lookup.get_template(_template)
    else: # Try looking inside basedir + '.tgz' tarball
        tar = tarfile.open(basedir + '.tgz')
        fileobj = tar.extractfile(_template)
        filedata = fileobj.read()
        return Template(filedata, lookup=template_lookup)

def plugin_template(plugin, template_path):
    """Returns a mako template object given the path to the template inside a plugin's directory or zip archive.
    Note: If the path doesn't exist on the filesystem this method will attempt to get the file out of the plugin's zip archive.

    Example:
        mytemplate = plugin_template('admin', templates/users_and_groups.mako)"""
    filepath = "plugins_enabled/%s/%s" % (plugin.__module__, template_path)
    template_lookup = TemplateLookup(directories=["html"], input_encoding='utf-8', output_encoding='utf-8', collection_size=1)
    if os.path.exists(filepath):
        return Template(filename=filepath, lookup=template_lookup)
    else: # Try looking inside the plugin's zip archive
        try:
            template = Template(sys.modules[plugin.__module__].__loader__.get_data(template_path), lookup=template_lookup)
            return template
        except: # Couldn't find the template there either
            return "Template, '%s' could not be found on the filesystem or in the plugin's zip file." % template_path

def get_groups(user=None):
    """Returns a list of groups the given user is a member of.
    If no user is provided, return a list of all groups."""
    group_list = []
    if not user:
        f = open('/etc/group', 'r')
        group_file = f.readlines()
        for line in group_file:
            line_list = line.split(':')
            group_list.append(line_list[0])
        return group_list
    groups = Popen(["groups", user], stdout=PIPE).communicate()[0]
    groups = groups.rstrip('\n').split(' ')
    if "No such user" in groups:
        return None
    else:
        return groups

# TODO:  Find out if I can get rid of these
#def get_valid_users():
    #"""Returns a list of users in /etc/shadow that have a password set"""
    #user_list = []
    #if os.path.exists('/etc/shadow'):
        #password_file = '/etc/shadow'
    #else:
        #password_file = '/etc/passwd'
    #shadow = open(password_file, 'r').readlines()
    #for line in shadow:
        #cols = line.split(':')
        #if len(cols[1]) > 4: # This avoids things like "!!" and "*LK*"
            #user_list.append(cols[0])
    #return user_list

#def get_valid_groups():
    #"""Returns a list of groups that valid users belong to"""
    #group_list = []
    #users = get_valid_users()
    #for user in users:
        #groups = get_groups(user)
        #for group in groups:
            #if group not in group_list:
                #group_list.append(group)
    #return group_list

class LazyProperty(object):
    """Decorator to make class methods return attributes only when accessed"""
    def __init__(self, func):
        self._func = func
        self.__name__ = func.__name__
        self.__doc__ = func.__doc__

    def __get__(self, obj, klass=None):
        if obj is None: return None
        result = obj.__dict__[self.__name__] = self._func(obj)
        return result

def split_seq(seq,size):
    """ Split up seq in pieces of size """
    return [seq[i:i+size] for i in range(0, len(seq), size)]

def sortby(_list, n):
    _list.sort(lambda x,y:cmp(x[n], y[n]))
    return _list

def sortby_field(sort_field, fields, _list):
    """Sorts a list of lists or tuples ('_list') by 'sort_field' given a list of 'fields' that maps to the sublists' indexes (if that makes sense?)"""
    out_list = []
    for field in fields:
        if sort_field == field:
            out_list = sortby(_list, fields.index(field))
    return out_list

def format_columns(string_list, columns, separator=" "):
    """A generator to produce equal-width columns from string_list

    Thanks to ΤΖΩΤΖΙΟΥ (http://stackoverflow.com/users/6899/)
    """
    import itertools
    sublists = []

    # empty_str based on item 0 of string_list
    try:
        empty_str = type(string_list[0])()
    except IndexError: # string_list is empty
        return

    # create a sublist for every column
    for column in xrange(columns):
        sublists.append(string_list[column::columns])

    # find maximum length of a column
    max_sublist_len = max(itertools.imap(len, sublists))

    # make all columns same length
    for sublist in sublists:
         if len(sublist) < max_sublist_len:
             sublist.append(empty_str)

    # calculate a format string for the output lines
    format_str = separator.join(
        "%%-%ds" % max(itertools.imap(len, sublist))
         for sublist in sublists)

    for line_items in itertools.izip(*sublists):
        yield format_str % line_items

def current_user():
    """Returns the name of the current logged in user"""
    return cherrypy.request.login

def jqgrid_format(header, list_data, id_field_index=None):
    """Convert a list of lists into a format that jqGrid will like when it is rendered via JSON.
    If an 'id_field_index' is specified, that field will be used as the 'id' JSON parameter."""
    out_list = []
    for row in list_data:
        if id_field_index:
            key = row[id_field_index]
        else:
            key = row[0]
        # BE AWARE: Dots and spaces are not allowed in element IDs
        cells = []
        for item in row:
            cells.append(item)
        out_list.append({'id': key, 'cell': cells})
    return out_list

def jqgrid_select_options(_list):
    """Turns a list of items into a format compatible with jqGrid selection lists.

    Here's an example of how jqgrid expects selection options:
        {value:"FE:FedEx;IN:InTime;TN:TNT;AR:ARAMEX"}

    This function just takes the items in the given list and doubles them up like so:
        'item1:item1;item2:item2;'"""
    values = ""
    for item in _list:
        values += "%s:%s;" % (item, item)
    values = values.rstrip(';') # Remove the trailing semicolon
    return values

def search_cells(header, _list, search_string, search_field, search_type):
    """Given a tuple of lists (cells) and a header (matching the cells), returns a list of items that match the given search_string, search_field, and search_type.

    Valid search types:
    eq - Equal to
    ne - Not equal to
    bw - Begins with
    ew - Ends with
    lt - Less than
    le - Less than or equal to
    gt - Greater than
    ge - Greater than or equal to
    cn - Contains"""
    out_list = []
    for item in _list:
        # ['zeroinst', 'Disabled', 111, 121, '', '/home/zeroinst', '/bin/false']
        if search_type == "bw": # Begins with
            if str(item[header.index(search_field)]).startswith(search_string):
                out_list.append(item)
        elif search_type == "ew":
            if str(item[header.index(search_field)]).endswith(search_string):
                out_list.append(item)
        elif search_type == "eq":
            if str(item[header.index(search_field)]) == search_string:
                out_list.append(item)
        elif search_type == "ne":
            if str(item[header.index(search_field)]) != search_string:
                out_list.append(item)
        elif search_type == "lt":
            if item[header.index(search_field)] < int(search_string):
                out_list.append(item)
        elif search_type == "le":
            if item[header.index(search_field)] <= int(search_string):
                out_list.append(item)
        elif search_type == "gt":
            if item[header.index(search_field)] > int(search_string):
                out_list.append(item)
        elif search_type == "ge":
            if item[header.index(search_field)] >= int(search_string):
                out_list.append(item)
        elif search_type == "cn":
            if search_string in item[header.index(search_field)]:
                out_list.append(item)
    return out_list

def jqgrid_json(self, cell_list, header, id_field_index=None, rows=None, sidx=None, _search=False, searchField=None, searchOper=None, searchString=None, page=None, sord=None):
    """Returns a list of cells encoded in a format jqgrid likes.

    Required variables:
    @header: The column names (must match cell_list).  Example:
        header = ['primary_key', 'firstname', 'lastname']
    @cell_list: A list containing lists of cells.  Example:
        cell_list = [
            ['1', 'John', 'Smith'],
            ['2', 'Bob', 'Jones']
        ]

    Optional variables:
    @id_field_index: The index of the field to use as jqGrid's 'id' parameter.
    @rows: Number of rows to return.  If this is not set all rows will be returned.
    @sidx: Field to sort the output by.
    @sord: Either 'asc' (ascending) or 'desc' (descending).
    @_search: True or False (whether or not this is a search).  Default is False.  Also takes strings of 'true' and 'false'.
    @searchField: The field we're searching on.
    @searchOper: The type of search (eq = equals, cn = contains, etc--see the docstring of search_cells for all the operators).
    @searchString: The string to match for our search.
    @page: The page number to return (if paginating)."""
    if not page:
        page = 1
    if sidx:
        cell_list = sortby_field(sidx, header, cell_list) # Sort the list by field
        if sord == 'desc':
            cell_list.reverse()
    if _search == "true" or _search is True:
        cell_list = search_cells(header, cell_list, searchString, searchField, searchOper)
    line_count = len(cell_list)
    try:
        jqgrid_cells = jqgrid_format(header, cell_list, id_field_index)
    except IndexError: # There was a problem with the cell_list
        jqgrid_cells = {}
    if rows:
        paginated_list = split_seq(jqgrid_cells,int(rows)) # Divide up the results into chunks of 'rows' size
        total_pages = len(paginated_list)
        if total_pages == 0:
            jqgrid_dict = { # No records to return
                'total': total_pages, # This will be 0
                'records': line_count, # Ditto
                'page': page, # Always going to be 1
                'rows': ''}
        else:
            paginated_list.insert(0,'') # Make the first item empty so page 1 equals list[1] (for convenience)
            jqgrid_dict = {
                'total': total_pages,
                'records': line_count,
                'page': page,
                'rows': paginated_list[int(page)]}
    else: # Return all records (don't paginate)
        jqgrid_dict = {
        'total': line_count,
        'records': line_count,
        'page': page,
        'rows': jqgrid_cells}
    return json.dumps(jqgrid_dict)

def jquery_ui_error_msg(message):
    """Returns 'message' surrounded by jquery-ui divs indicating an error (makes it look nice)"""
    return '<div class="ui-state-error"><span id="close_this" class="ui-icon ui-icon-closethick" style="float: right; cursor: pointer;" onclick="$(this).parent().parent().hide();clearTimeout(resultTimeout);return false;"></span><span class="ui-icon ui-icon-alert ui-icon-left" style="margin-right: .3em;"></span><strong>Error:</strong> %s</div>' % message

def jquery_ui_highlight_msg(message):
    """Returns 'message' surrounded by jquery-ui divs indicating an informational message (makes it look nice)"""
    return '<div class="ui-state-highlight"><span id="close_this" class="ui-icon ui-icon-closethick" style="float: right; cursor: pointer;" onclick="$(this).parent().parent().hide();clearTimeout(resultTimeout);return false;"></span><span class="ui-icon ui-icon-check ui-icon-left" style="margin-right: .3em;"></span> %s</div>' % message

def timeout(func, args=(), kwargs={}, timeout_duration=10, default=None):
    """This function will spawn a thread and run the given function
    using the args, kwargs and return the given default value if the
    timeout_duration is exceeded.

    Invoking it with a 5 second timeout:
        result = timeout(remote_calculate, (myarg,), timeout_duration=5)

    (taken from http://stackoverflow.com/questions/492519/timeout-on-a-python-function-call)
    """
    import threading
    class InterruptableThread(threading.Thread):
        def __init__(self):
            threading.Thread.__init__(self)
            self.result = default
        def run(self):
            self.result = func(*args, **kwargs)
    it = InterruptableThread()
    it.start()
    it.join(timeout_duration)
    if it.isAlive():
        return it.result
    else:
        return it.result

def service_control(service, operation):
    """
    This function just runs /etc/init.d/<service> <operation>
    Returns True if the operation was successful
    Returns the command output on failure
    """
    from commands import getstatusoutput
    retcode, output = getstatusoutput('/etc/init.d/%s %s' % (service, operation))
    if retcode == 0:
        return True
    else:
        return output

class Flot(object):
    """A class to store and return data to be plotted by flot.
    @data: A list representing list pairs of x and y data points.
        Example: [[0, 3], [10, 3]]
        That is: [[x, y], [x, y]]... and so on
    @options:  A dictionary of options to be passed to flot.
        Valid options are:
        {
            color: color or number
            label: string
            lines: specific lines options
            bars: specific bars options
            points: specific points options
            xaxis: 1 or 2
            yaxis: 1 or 2
            clickable: boolean
            hoverable: boolean
            shadowSize: number
        }"""
    params = ["data", "options"]

    def __init__(self, data, options={}):
        self.data = data
        self.options = options

    def to_json(self):
        """Returns our variables in a json-encoded format that flot likes"""
        flot_dict = {'data': self.data}
        for k, v in self.options.iteritems():
            flot_dict.update({k:v})
        return json.dumps(flot_dict)

    def __str__(self):
        return self.to_json()